/**
 * about.c
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <windows.h>
#include <windowsx.h>

#include "util.h"
#include "JanusHS.h"
#include "resource.h"

typedef struct _INPUTBOX_PARAM {
  LPCSTR pPrompt;
  LPCSTR pTitle;
  PSTR pBuff;
  UINT uMax;
  DWORD dwStyle;
  DWORD dwErr;
} INPUTBOX_PARAM, *PINPUTBOX_PARAM;

#define MAXINPUT 256

extern HINSTANCE g_hInstance;

/*------------------------------------------------------------*/
static BOOL
inputbox_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  DWORD dwSize = 0, dwLang = 0, dwErr = ERROR_SUCCESS, dwHandle = 0;
  CHAR szFilename[MAX_PATH + 1];
  LPVOID pBuff = NULL;
  LPSTR buff = NULL;
  CHAR szProductName[MAX_PATH + 1] = "", szProductVersion[MAX_PATH + 1] = "", szLegalCopyright[MAX_PATH + 1] = "";
  UINT uLen = 0, uLabelLen = 0;
  CHAR szLabel[MAX_PATH * 3 + 1] = "";
  
  if (GetModuleFileName(NULL, szFilename, MAX_PATH) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if ((dwSize = GetFileVersionInfoSize(szFilename, NULL)) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  pBuff = (PBYTE)MALLOC(dwSize);
  if (!pBuff) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ZeroMemory(pBuff, dwSize);
  if (GetFileVersionInfo(szFilename, dwHandle, dwSize, pBuff) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if (VerQueryValue(pBuff, "\\StringFileInfo\\041104b0\\ProductName", (LPVOID *)&buff, &uLen) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if (uLen > 0) {
    ZeroMemory(szProductName, MAX_PATH);
    STRNCPY(szProductName, buff, uLen);
  }
  if (VerQueryValue(pBuff, "\\StringFileInfo\\041104B0\\ProductVersion", (LPVOID *)&buff, &uLen) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if (uLen > 0) {
    ZeroMemory(szProductVersion, MAX_PATH);
    STRNCPY(szProductVersion, buff, uLen);
  }
  if (VerQueryValue(pBuff, "\\StringFileInfo\\041104B0\\LegalCopyright", (LPVOID *)&buff, &uLen) == 0) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if (uLen > 0) {
    ZeroMemory(szLegalCopyright, MAX_PATH);
    STRNCPY(szLegalCopyright, buff, uLen);
  }
  wsprintf(szLabel, "%s Version %s", szProductName, szProductVersion);
  Static_SetText(GetDlgItem(hwnd, IDC_STATIC_PRODUCTNAME), szLabel);
  Static_SetText(GetDlgItem(hwnd, IDC_STATIC_LEGALCOPYRIGHT), szLegalCopyright);
finish_func:
  if (pBuff) FREE(pBuff);
  if (dwErr != ERROR_SUCCESS) EndDialog(hwnd, (int)dwErr);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
inputbox_Cls_OnClose(HWND hwnd) {
  EndDialog(hwnd, (INT)ERROR_SUCCESS);
}

/*------------------------------------------------------------*/
static VOID
inputbox_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  }
}

/*------------------------------------------------------------*/
BOOL CALLBACK
procAboutDialog(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_MSG(hDlg, WM_INITDIALOG, inputbox_Cls_OnInitDialog);
    HANDLE_MSG(hDlg, WM_CLOSE, inputbox_Cls_OnClose);
    HANDLE_MSG(hDlg, WM_COMMAND, inputbox_Cls_OnCommand);
  default:
    return FALSE;
  }
  return TRUE;
}

/*------------------------------------------------------------*/
DWORD
ShowAboutDialog(HWND hwnd) {
  INT iRet;
  iRet = DialogBox(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_VERSIONINFO), hwnd, procAboutDialog);
  return (DWORD) iRet;
}

/*------------------------------------------------------------*/
