/**
 * chainedprocess.c
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include "util.h"
#include "gpgez.h"
#include "JanusHS.h"
#include "resource.h"

extern HINSTANCE g_hInstance;
extern CRITICAL_SECTION cs;

typedef struct _CHAINEDPROCESS {
  PCHAINEDPROCESS_NODE phead;
  PHANDLE phRead;
  PHANDLE phWrite;
  HWND hwndmsg;
} CHAINEDPROCESS, *PCHAINEDPROCESS;

/*------------------------------------------------------------*/
/**
 * \fn DWORD CreateChainedProcess(HWND hwnd, PCHAINEDPROCESS pcp)
 *
 * pCv쐬ȂÃvZXNAĎAI֐
 *
 */
static DWORD
CreateChainedProcess(HWND hwnd, PCHAINEDPROCESS pcp) {
  PCHAINEDPROCESS_NODE p = pcp->phead;
  DWORD dwObserve = 0, dwAvail = 0, dwEffective = 0;
  DWORD dwErr = ERROR_SUCCESS;
  SECURITY_ATTRIBUTES sa;
  BOOL fEffective = TRUE;
  UINT lcnt = 0;

  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;

  /* vZXԃpCvݒ肵ȂvZXN */
  EnterCriticalSection(&cs);
  dwObserve = 0;
  for (p = pcp->phead; p; p = p->next) {
    if (p == pcp->phead) {
      p->hRead = *(pcp->phRead);
    }
    if (p->next) {
      if (!CreatePipe(&(p->next->hRead), &(p->hWrite), &sa, 0)) {
        dwErr = GetLastError();
        DEBUGERROR(0, dwErr);
        LeaveCriticalSection(&cs);
        goto finish_func;
      }
    } else {
      p->hWrite = *(pcp->phWrite);
    }
    if ((dwErr = p->dwStartUp(p)) != ERROR_SUCCESS) {
      DEBUGERROR(0, dwErr);
      LeaveCriticalSection(&cs);
      goto finish_func;
    }
    if (p != pcp->phead) CloseHandleClean(&(p->hRead));
    if (p->next) CloseHandleClean(&(p->hWrite));
    if ((p->phObserve) && (*(p->phObserve) != INVALID_HANDLE_VALUE)) {
      dwObserve++;
    }
  }

  if (*(pcp->phRead) != INVALID_HANDLE_VALUE) CloseHandleClean(pcp->phRead);
  if (*(pcp->phWrite) != INVALID_HANDLE_VALUE) CloseHandleClean(pcp->phWrite);
  LeaveCriticalSection(&cs);

  /* evZX̏ԂĎ */
  while (dwObserve > 0) {
    fEffective = FALSE;
    for (p = pcp->phead; p; p = p->next) {
      if ((p->phObserve) && (*(p->phObserve) != INVALID_HANDLE_VALUE)) {
        do {
          if (!GetExitCodeProcess(p->hProcess, &(p->dwExitCode))) {
            dwErr = GetLastError();
            DEBUGERROR(0, dwErr);
            goto finish_func;
          }
          if (!PeekNamedPipe(*(p->phObserve), NULL, 0, NULL, &dwAvail, NULL)) {
            dwErr = GetLastError();
            if ((dwErr == ERROR_HANDLE_EOF) || (dwErr == ERROR_BROKEN_PIPE)) {
              dwErr = ERROR_SUCCESS;
              CloseHandleClean(p->phObserve);
              dwObserve--;
              break;
            } else {
              DEBUGERROR(0, dwErr);
              goto finish_func;
            }
          }
          if (dwAvail > 0) {
            fEffective = TRUE;
            if ((dwErr = p->dwProcObserved(p, pcp->hwndmsg)) != ERROR_SUCCESS) {
              DEBUGERROR(0, dwErr);
              goto finish_func;
            }
          } else if (p->dwExitCode != STILL_ACTIVE) {
            CloseHandleClean(p->phObserve);
            dwObserve--;
            break;
          }
        } while (dwAvail > 0);
      }
    }
    if (!fEffective) {
      Sleep(100);
      lcnt++;
      if (lcnt > 600) {
        if (MessageBox(hwnd, "Time Wait Exceeds 1 minute.\nCoutinue ?", "Confirmation", MB_YESNO) == IDNO) {
          SetLastError(ERROR_CANCELLED);
          goto finish_func;
        }
        lcnt = 0;
      }
    }
  }

finish_func:
  /* evZX̏I */

  EnterCriticalSection(&cs);
  for (p = pcp->phead; p; p = p->next) {
    if (p != pcp->phead) {if (p->hRead != INVALID_HANDLE_VALUE) CloseHandleClean(&(p->hRead));}
    if (p->next) {if (p->hWrite != INVALID_HANDLE_VALUE) CloseHandleClean(&(p->hWrite));}
    if (p->hProcess != INVALID_HANDLE_VALUE) {
      WaitForSingleObject(p->hProcess, INFINITE);
      p->dwFinishUp(p);
      CloseHandleClean(&(p->hProcess));
    }
  }
  LeaveCriticalSection(&cs);
  if (dwErr == ERROR_SUCCESS) {
    Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
    Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  }
  return dwErr;
}

/*------------------------------------------------------------*/
/**
 * \fn BOOL chainedprocess_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam)
 *
 * _CAȌ
 * bZ[W\GfBbg[hI[Ŏgpɂ
 * OK{^̓vZXI܂ŎgpsALZ{^̂ݎgpƂ
 *
 */
static BOOL
chainedprocess_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  PCHAINEDPROCESS pcp;

  pcp = (PCHAINEDPROCESS)lParam;
  pcp->hwndmsg = GetDlgItem(hwnd, IDC_EDIT_MESSAGE1);
  SetWindowLong(hwnd, GWL_USERDATA, (LONG)pcp);
  Edit_Enable(pcp->hwndmsg, TRUE);
  Edit_SetText(pcp->hwndmsg, "");
  Edit_SetReadOnly(pcp->hwndmsg, TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);

  PostMessage(hwnd, WM_CREATEPROCESS, 0, (LONG)pcp);

  return TRUE;
}

/*------------------------------------------------------------*/
/**
 * \fn VOID chainedprocess_Cls_OnClose(HWND hwnd)
 *
 * _CAȌI
 * ̃_CAOŋNꂽvZXΏI
 *
 */
static VOID
chainedprocess_Cls_OnClose(HWND hwnd) {
  PCHAINEDPROCESS pcp;
  PCHAINEDPROCESS_NODE p;
  DWORD dwStatus;

  EnterCriticalSection(&cs);
  pcp = (PCHAINEDPROCESS)GetWindowLong(hwnd, GWL_USERDATA);
  if (*(pcp->phRead) != INVALID_HANDLE_VALUE) CloseHandleClean(pcp->phRead);
  if (*(pcp->phWrite) != INVALID_HANDLE_VALUE) CloseHandleClean(pcp->phWrite);
  for(p = pcp->phead; p; p = p->next) {
    if (p->hProcess != INVALID_HANDLE_VALUE) {
      if (GetExitCodeProcess(p->hProcess, &dwStatus)) {
        if (dwStatus == STILL_ACTIVE) {
          TerminateProcess(p->hProcess, 0);
          p->dwFinishUp(p);
          if (p != pcp->phead) {if (p->hRead != INVALID_HANDLE_VALUE) CloseHandleClean(&(p->hRead));}
          if (p->next) {if (p->hWrite != INVALID_HANDLE_VALUE) CloseHandleClean(&(p->hWrite));}
          if (p->hProcess != INVALID_HANDLE_VALUE) CloseHandleClean(&(p->hProcess));
        }
      }
    }
  }
  LeaveCriticalSection(&cs);

  EndDialog(hwnd, (INT)ERROR_CANCELLED);
  return;
}

/*------------------------------------------------------------*/
/**
 * \fn VOID chainedprocess_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
 *
 * _CAÕ{^NbNꂽꍇ̏
 * OKƃLẐ
 *
 */
static VOID
chainedprocess_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  case IDCANCEL:
    chainedprocess_Cls_OnClose(hwnd);
    break;
  }
}

/*------------------------------------------------------------*/
/**
 * \fn BOOL CALLBACK procDlgChainedProcess(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
 *
 * _CAÕR[obN֐
 *
 */
BOOL CALLBACK
procDlgChainedProcess(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_MSG(hDlg, WM_INITDIALOG, chainedprocess_Cls_OnInitDialog);
    HANDLE_MSG(hDlg, WM_CLOSE, chainedprocess_Cls_OnClose);
    HANDLE_MSG(hDlg, WM_COMMAND, chainedprocess_Cls_OnCommand);
  case WM_CREATEPROCESS:
    CreateChainedProcess(hDlg, (PCHAINEDPROCESS)lParam);
    return TRUE;
  default:
    return FALSE;
  }
  return TRUE;
}

/*------------------------------------------------------------*/
/**
 * \fn DWORD raiseupChainedProcess(HWND hwnd, PCHAINEDPROCESS_NODE pnode_head, PHANDLE phRead, PHANDLE phWrite)
 *
 * bZ[W\p̃_CAO\ŁApCvłȂꂽvZXAċN֐
 *
 */
DWORD
raiseupChainedProcess(HWND hwnd, PCHAINEDPROCESS_NODE pnode_head, PHANDLE phRead, PHANDLE phWrite) {
  CHAINEDPROCESS cp;
  DWORD dwErr = ERROR_SUCCESS;

  cp.phead = pnode_head;
  cp.phRead = phRead;
  cp.phWrite = phWrite;
  dwErr = (DWORD)DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROCESS), hwnd, procDlgChainedProcess, (LPARAM)&cp);
  return dwErr;
}

/*------------------------------------------------------------*/
