/**
 * editKeyRing.c
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include "util.h"
#include "gpgez.h"
#include "JanusHS.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

typedef struct _JANUS_EDITKEY {
  LPCSTR szGpgHome;
  PSTRLIST pslPubRing;
  PSTRLIST pslSecRing;
} JANUS_EDITKEY, *PJANUS_EDITKEY;

/*------------------------------------------------------------*/
static DWORD
editkeyring_show(HWND hwnd) {
  HANDLE hReadP = INVALID_HANDLE_VALUE;
  HANDLE hStdoutC = INVALID_HANDLE_VALUE;
  HANDLE hProcess = INVALID_HANDLE_VALUE;
  CHAR szLine[MAX_LINE + 1];
  CHAR szUid[MAX_LINE + 1];
  CHAR szFpr[MAX_LINE + 1];
  UINT i;
  LPSTR ps, pe;
  LV_ITEM item;
  LPSTR psjis = NULL;
  UINT iLine = 0;
  DWORD dwErr = ERROR_SUCCESS;
  HWND hwndList;
  PJANUS_EDITKEY pjek;

  pjek = (PJANUS_EDITKEY)GetWindowLong(hwnd, GWL_USERDATA);

  hwndList = GetDlgItem(hwnd, IDC_LIST_KEYS);
  szUid[0] = '\0';
  szFpr[0] = '\0';
  if (!ListView_DeleteAllItems(hwndList)) {
    dwErr = GetLastError();
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }

  dwErr = CreateReadPipe(&hReadP, &hStdoutC);
  if (dwErr != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  dwErr = gpgez_listkeys(pjek->szGpgHome, pjek->pslPubRing, FALSE, hStdoutC, &hProcess);
  CloseHandleClean(&hStdoutC);

  while (dwErr == ERROR_SUCCESS) {
    dwErr = ReadPipeLine(&hReadP, szLine, MAX_LINE);
    if ((dwErr != ERROR_SUCCESS) && (dwErr != ERROR_HANDLE_EOF)) {
      DEBUGERROR(0, dwErr);
      break;
    }
    if (!STRNCMPL("pub:", szLine)) {
      if (szUid[0] || szFpr[0]) {
        /* ListView ɒǉ */
        item.mask = LVIF_TEXT;
        item.pszText = szUid;
        item.iItem = iLine;
        item.iSubItem = 0;
        ListView_InsertItem(hwndList, &item);

        item.pszText = szFpr;
        item.iItem = iLine;
        item.iSubItem = 1;
        ListView_SetItem(hwndList, &item);

        szUid[0] = '\0';
        szFpr[0] = '\0';
        iLine++;
      }
      ps = szLine;
      for (i = 0; i < 9; i++) {
        if (!(ps = STRCHR(ps, ':'))) {
          dwErr = ERROR_INVALID_DATA;
          DEBUGERROR(0, dwErr);
          goto finish_func;
        }
        ps++;
      }
      pe = STRCHR(ps, ':');
      STRNCPY(szUid, ps, pe - ps);
      szUid[pe - ps] = '\0';
      if (!(psjis = utf82sjis(szUid))) {
        dwErr = ERROR_OUTOFMEMORY;
        DEBUGERROR(0, dwErr);
        break;
      }
      STRCPY(szUid, psjis);
      FREE(psjis);
    }
    if ((!STRNCMPL("fpr:", szLine)) && !szFpr[0]) {
      ps = szLine;
      for (i = 0; i < 9; i++) {
        if (!(ps = STRCHR(ps, ':'))) {
          dwErr = ERROR_INVALID_DATA;
          DEBUGERROR(0, dwErr);
          goto finish_func;
        }
        ps++;
      }
      pe = STRCHR(ps, ':');
      STRNCPY(szFpr, ps, pe - ps);
      szFpr[pe - ps] = '\0';
    }
    if (dwErr == ERROR_HANDLE_EOF) {
      if (szFpr[0]) {
        item.mask = LVIF_TEXT;
        item.pszText = szUid;
        item.iItem = iLine;
        item.iSubItem = 0;
        ListView_InsertItem(hwndList, &item);

        item.pszText = szFpr;
        item.iItem = iLine;
        item.iSubItem = 1;
        ListView_SetItem(hwndList, &item);

        szUid[0] = '\0';
        szFpr[0] = '\0';
        iLine++;
      }
      dwErr = ERROR_SUCCESS;
      break;
    }
  }
finish_func:
  CloseHandleClean(&hProcess);
  CloseHandleClean(&hReadP);
  CloseHandleClean(&hStdoutC);
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
editkeyring_importfile(HWND hwnd, LPCSTR pConfDir, PSTRLIST pSecKey, PSTRLIST pPubKey, LPCSTR pFilename) {
  PCHAINEDPROCESS_NODE phead = NULL, pp = NULL, pp1, ppt;
  GPGIMPORT_PARAM gparam;
  SECURITY_ATTRIBUTES sa;
  HANDLE hRead = INVALID_HANDLE_VALUE, hWrite = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pp1 = (PCHAINEDPROCESS_NODE)MALLOC(sizeof(CHAINEDPROCESS_NODE)))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  initializeGpgImportParam(&gparam, pConfDir, pSecKey, pPubKey);
  pp1->buff = (PVOID)&gparam;
  pp1->dwStartUp = startupGpgImport;
  pp1->dwProcObserved = procObservedGpgImport;
  pp1->dwFinishUp = finishupGpgImport;
  pp1->phObserve = NULL;
  pp1->next = NULL;
  phead = pp1;

  ZeroMemory(&sa, sizeof(sa));
  sa.nLength = sizeof(sa);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;
  if ((hRead = CreateFile(pFilename, GENERIC_READ, 0, &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) {
    dwErr = GetLastError();
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  hWrite = INVALID_HANDLE_VALUE;
  dwErr = raiseupChainedProcess(hwnd, phead, &hRead, &hWrite);
  if (dwErr != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
  }

finish_func:
  pp1 = phead;
  while (pp1) {
    ppt = pp1->next;
    FREE(pp1);
    pp1 = ppt;
  }
  if (hRead != INVALID_HANDLE_VALUE) CloseHandleClean(&hRead);
  if (hWrite != INVALID_HANDLE_VALUE) CloseHandleClean(&hWrite);
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
editkeyring_import(HWND hwnd) {
  OPENFILENAME ofn;
  CHAR szFileName[MAX_PATH + 1] = "";
  DWORD dwErr = ERROR_SUCCESS;
  PJANUS_EDITKEY pjek;

  pjek = (PJANUS_EDITKEY)GetWindowLong(hwnd, GWL_USERDATA);

  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = "gpg(*.gpg)\0*.gpg\0All files(*.*)\0*.*\0\0";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST;
  ofn.lpstrDefExt = "gpg";
  if (GetOpenFileName(&ofn)) {
    dwErr = editkeyring_importfile(hwnd, pjek->szGpgHome, pjek->pslSecRing, pjek->pslPubRing, szFileName);
    if (dwErr != ERROR_SUCCESS) {
      DEBUGERROR(0, dwErr);
    }
    editkeyring_show(hwnd);
  } else {
    /* G[`FbN */
  }
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
editkeyring_export(HWND hwnd) {
  OPENFILENAME ofn;
  HANDLE hProcess = INVALID_HANDLE_VALUE;
  HANDLE hFile = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;
  DWORD dwExitCode;
  LVITEM item;
  HWND hwndKeyList;
  INT iItem = -1;
  LPSTR pUid = NULL;
  LPSTR pFpr = NULL;
  CHAR szFileTitle[MAX_PATH + 1] = "";
  CHAR szFileName[MAX_PATH + 1] = "";
  CHAR szLine[MAX_LINE + 1] = "";
  PJANUS_EDITKEY pjek;
  SECURITY_ATTRIBUTES sa;

  pjek = (PJANUS_EDITKEY)GetWindowLong(hwnd, GWL_USERDATA);

  item.mask = LVIF_TEXT;
  item.pszText = szLine;
  item.cchTextMax = MAX_LINE;
  hwndKeyList = GetDlgItem(hwnd, IDC_LIST_KEYS);
  iItem = ListView_GetNextItem(hwndKeyList, -1, LVNI_ALL | LVNI_SELECTED);
  if (iItem == -1) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_NOKEYSELECTED, m_buff, MAX_PATH) == 0) {
      STRCPY(m_buff, "No Key Selected");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK | MB_ICONEXCLAMATION);
    dwErr = ERROR_SUCCESS;
    goto finish_func;
  }
  if (ListView_GetNextItem(hwndKeyList, iItem, LVNI_ALL | LVNI_SELECTED) != -1) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_MORETHANONEKEYSELECTEDFOREXPORT, m_buff, MAX_PATH) == 0) {
      STRCPY(m_buff, "Please select one key to export");
    }
    MessageBox(hwnd, m_buff, "ERROR", MB_OK | MB_ICONEXCLAMATION);
    dwErr = ERROR_SUCCESS;
    goto finish_func;
  }
  item.iItem = iItem;
  item.iSubItem = 0;
  ListView_GetItem(hwndKeyList, &item);
  if (!(pUid = STRDUP(item.pszText))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  item.iSubItem = 1;
  ListView_GetItem(hwndKeyList, &item);
  if (!(pFpr = STRDUP(item.pszText))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }

  szFileTitle[0] = '\0';
  if (STRLEN(pUid) < MAX_PATH - 5) {
    STRCPY(szFileTitle, pUid);
  }
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = "gpg(*.gpg)\0*.gpg\0All files(*.*)\0*.*\0\0";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT;
  ofn.lpstrDefExt = "gpg";
  if (GetSaveFileName(&ofn)) {
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.lpSecurityDescriptor = NULL;
    sa.bInheritHandle = TRUE;
    hFile = CreateFile(ofn.lpstrFile, GENERIC_WRITE, 0, &sa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE) {
      dwErr = GetLastError();
      DEBUGERROR(0, dwErr);
      goto finish_func;
    }
    dwErr = gpgez_export(pjek->szGpgHome, pjek->pslPubRing, hFile, pFpr, &hProcess);
    if (dwErr != ERROR_SUCCESS) {
      DEBUGERROR(0, dwErr);
      goto finish_func;
    }
    WaitForSingleObject(hProcess, INFINITE);
    if (!GetExitCodeProcess(hProcess, &dwExitCode)) {
      dwErr = GetLastError();
      DEBUGERROR(0, dwErr);
    }
    CloseHandleClean(&hProcess);
    CloseHandleClean(&hFile);
  } else {
    /* G[`FbN */
  }
finish_func:
  CloseHandleClean(&hProcess);
  CloseHandleClean(&hFile);
  if (pUid) FREE(pUid);
  if (pFpr) FREE(pFpr);
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
editkeyring_deletefile(HWND hwnd, LPCSTR pConfDir, PSTRLIST pSecKey, PSTRLIST pPubKey, LPCSTR pFpr) {
  PCHAINEDPROCESS_NODE phead = NULL, pp = NULL, pp1, ppt;
  HANDLE hRead = INVALID_HANDLE_VALUE, hWrite = INVALID_HANDLE_VALUE;
  GPGDELETE_PARAM gparam;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pp1 = (PCHAINEDPROCESS_NODE)MALLOC(sizeof(CHAINEDPROCESS_NODE)))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  initializeGpgDeleteParam(&gparam, pConfDir, pSecKey, pPubKey, pFpr);
  pp1->buff = (PVOID)&gparam;
  pp1->dwStartUp = startupGpgDelete;
  pp1->dwProcObserved = procObservedGpgDelete;
  pp1->dwFinishUp = finishupGpgDelete;
  pp1->phObserve = NULL;
  pp1->next = NULL;
  phead = pp1;

  hRead = INVALID_HANDLE_VALUE;
  hWrite = INVALID_HANDLE_VALUE;
  dwErr = raiseupChainedProcess(hwnd, phead, &hRead, &hWrite);
  if (dwErr != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
  }

finish_func:
  pp1 = phead;
  while (pp1) {
    ppt = pp1->next;
    FREE(pp1);
    pp1 = ppt;
  }
  if (hRead != INVALID_HANDLE_VALUE) CloseHandleClean(&hRead);
  if (hWrite != INVALID_HANDLE_VALUE) CloseHandleClean(&hWrite);
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
editkeyring_delete(HWND hwnd) {
  DWORD dwErr = ERROR_SUCCESS;
  LVITEM item;
  HWND hwndKeyList;
  INT iItem = -1;
  LPSTR pFpr = NULL;
  PJANUS_EDITKEY pjek;
  CHAR szLine[MAX_LINE + 1];

  pjek = (PJANUS_EDITKEY)GetWindowLong(hwnd, GWL_USERDATA);

  /* IԂ̌𒲂ׂ */
  item.mask = LVIF_TEXT;
  item.pszText = szLine;
  item.cchTextMax = MAX_LINE;
  hwndKeyList = GetDlgItem(hwnd, IDC_LIST_KEYS);
  iItem = ListView_GetNextItem(hwndKeyList, -1, LVNI_ALL | LVNI_SELECTED);
  if (iItem == -1) {
    MessageBox(hwnd, "Please select a key to delete", "ERROR", MB_OK);
    dwErr = ERROR_SUCCESS;
    goto finish_func;
  }
  if (ListView_GetNextItem(hwndKeyList, iItem, LVNI_ALL | LVNI_SELECTED) != -1) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_MORETHANONEKEYSELECTEDFORDELETE, m_buff, MAX_PATH) == 0) {
      STRCPY(m_buff, "Please select one key to delete");
    }
    MessageBox(hwnd, m_buff, "ERROR", MB_OK);
    dwErr = ERROR_SUCCESS;
    goto finish_func;
  }
  item.iItem = iItem;
  item.iSubItem = 0;
  item.iSubItem = 1;
  ListView_GetItem(hwndKeyList, &item);
  if (!(pFpr = STRDUP(item.pszText))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }

  dwErr = editkeyring_deletefile(hwnd, pjek->szGpgHome, pjek->pslSecRing, pjek->pslPubRing, pFpr);
  if (dwErr != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
  }

  editkeyring_show(hwnd);

finish_func:
  if (pFpr) FREE(pFpr);
  return dwErr;
}

/*------------------------------------------------------------*/
static BOOL
editkeyring_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  LV_COLUMN lvcol;

  SetWindowLong(hwnd, GWL_USERDATA, (LONG)lParam);

  lvcol.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
  lvcol.fmt = LVCFMT_LEFT;
  lvcol.cx = 200;
  lvcol.pszText = "Name";
  lvcol.iSubItem = 0;
  ListView_InsertColumn(GetDlgItem(hwnd, IDC_LIST_KEYS), 0, &lvcol);

  lvcol.cx = 300;
  lvcol.pszText = "Fingerprint";
  lvcol.iSubItem = 1;
  ListView_InsertColumn(GetDlgItem(hwnd, IDC_LIST_KEYS), 1, &lvcol);

  return TRUE;
}

/*------------------------------------------------------------*/
static void
editkeyring_Cls_OnShowWindow(HWND hwnd, BOOL fShow, UINT status) {
  if (fShow) {
    editkeyring_show(hwnd);
  }
}

/*------------------------------------------------------------*/
static VOID
editkeyring_Cls_OnClose(HWND hwnd) {
  EndDialog(hwnd, (INT)ERROR_CANCELLED);
  return;
}

/*------------------------------------------------------------*/
static VOID
editkeyring_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  case IDC_BUTTON_IMPORTKEY:
    editkeyring_import(hwnd);
    break;
  case IDC_BUTTON_EXPORTKEY:
    editkeyring_export(hwnd);
    break;
  case IDC_BUTTON_DELETEKEY:
    editkeyring_delete(hwnd);
    break;
  }
}

/*------------------------------------------------------------*/
BOOL CALLBACK
procDlgEditKey(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_MSG(hDlg, WM_INITDIALOG, editkeyring_Cls_OnInitDialog);
    HANDLE_MSG(hDlg, WM_SHOWWINDOW, editkeyring_Cls_OnShowWindow);
    HANDLE_MSG(hDlg, WM_CLOSE, editkeyring_Cls_OnClose);
    HANDLE_MSG(hDlg, WM_COMMAND, editkeyring_Cls_OnCommand);
  default:
    return FALSE;
  }
  return TRUE;
}

/*------------------------------------------------------------*/
DWORD editGpgKey(HWND hwnd, LPCSTR szGpgHome, PSTRLIST pslSecRing, PSTRLIST pslPubRing) {
  JANUS_EDITKEY jkey;
  DWORD dwErr = ERROR_SUCCESS;

  jkey.szGpgHome = szGpgHome;
  jkey.pslSecRing = pslSecRing;
  jkey.pslPubRing = pslPubRing;
  dwErr = (DWORD)DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_EDIT_KEYRING), hwnd, procDlgEditKey, (LPARAM)&jkey);

  return dwErr;
}

/*------------------------------------------------------------*/
