/**
 * generateKey.c
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <windows.h>
#include <windowsx.h>
#include "JanusHS.h"
#include "gpgez.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

typedef struct param_genkey {
  LPCSTR szGpgHome; /* gpg.exe--homedirɎw肷镶A\͊܂܂Ȃ */
  PSTRLIST pslSecRing;
  PSTRLIST pslPubRing;
  LPCSTR pName;
  LPCSTR pEmail;
  LPCSTR pComment;
  LPCSTR pPassphrase;
  LPSTR pFpr;
  HWND hwnd;
  HWND hwndMsg;
} PARAM_GENKEY, *PPARAM_GENKEY;

/*------------------------------------------------------------*/
static DWORD
WriteFileLine1(HANDLE hFile, LPCSTR s1) {
  DWORD dwWritten;
  if (!WriteFile(hFile, s1, STRLEN(s1), &dwWritten, NULL)) return GetLastError();
  if (!WriteFile(hFile, CRLF, STRLEN(CRLF), &dwWritten, NULL)) return GetLastError();
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD
WriteFileLine2(HANDLE hFile, LPCSTR s1, LPCSTR s2) {
  DWORD dwWritten;
  if (!WriteFile(hFile, s1, STRLEN(s1), &dwWritten, NULL)) return GetLastError();
  if (!WriteFile(hFile, " ", STRLEN(" "), &dwWritten, NULL)) return GetLastError();
  if (!WriteFile(hFile, s2, STRLEN(s2), &dwWritten, NULL)) return GetLastError();
  if (!WriteFile(hFile, CRLF, STRLEN(CRLF), &dwWritten, NULL)) return GetLastError();
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------*/
static DWORD
validateName(LPCSTR szName) {
  LPCSTR pSrc;
  DWORD dwErr = ERROR_SUCCESS;

  if (STRLEN(szName) < 5) {
    dwErr = ERROR_INVALID_DATA;
    goto finish_func;
  }
  if ((*szName >= '0') && (*szName <= '9')) {
    dwErr = ERROR_INVALID_DATA;
    goto finish_func;
  }
  for (pSrc = szName; *pSrc; pSrc++) {
    if ((*pSrc == '<') || (*pSrc == '>')) {
      dwErr = ERROR_INVALID_DATA;
      goto finish_func;
    }
  }
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
validateEmail(LPCSTR szEmail) {
  DWORD dwAt;
  LPCSTR pSrc;
  DWORD dwErr = ERROR_SUCCESS;

  if ((*szEmail == '@') ||
      (*(szEmail + strlen(szEmail) - 1) == '@') ||
      (*(szEmail + strlen(szEmail) - 1) == '.') ||
      (strstr(szEmail, ".."))) {
    dwErr = ERROR_INVALID_DATA;
    goto finish_func;
  }
  for (dwAt = 0, pSrc = szEmail; *pSrc; pSrc++) {
    if (*pSrc == '@') dwAt++;
  }
  if (dwAt != 1) {
    dwErr = ERROR_INVALID_DATA;
    goto finish_func;
  }
  for (dwAt = 0, pSrc = szEmail; *pSrc; pSrc++) {
    if (*pSrc & 0x80) {
      dwErr = ERROR_INVALID_DATA;
      goto finish_func;
    }
    if (*pSrc == '@') {
      dwAt++;
    } else {
      if ((dwAt == 0) &&
          (!isalnum(*pSrc)) &&
          (*pSrc != '+') &&
          (*pSrc != '-') &&
          (*pSrc != '_') &&
          (*pSrc != '.')) {
        dwErr = ERROR_INVALID_DATA;
        goto finish_func;
      } else if ((dwAt == 1)  &&
                 (!isalnum(*pSrc)) &&
                 (*pSrc != '+') &&
                 (*pSrc != '-') &&
                 (*pSrc != '_') &&
                 (*pSrc != '.')) {
        dwErr = ERROR_INVALID_DATA;
        goto finish_func;
      }
    }
  }
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
validateComment(LPCSTR szComment) {
  LPCSTR pSrc;
  DWORD dwErr = ERROR_SUCCESS;

  for (pSrc = szComment; *pSrc; pSrc++) {
    if ((*pSrc == '(') || (*pSrc == ')')) {
      dwErr = ERROR_INVALID_DATA;
      goto finish_func;
    }
  }
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
genkey_dwGenerateGPGKeys2(PPARAM_GENKEY pgen) {
  HANDLE hCommandC = INVALID_HANDLE_VALUE;
  HANDLE hCommandP = INVALID_HANDLE_VALUE;
  HANDLE hStatusC = INVALID_HANDLE_VALUE;
  HANDLE hStatusP = INVALID_HANDLE_VALUE;
  HANDLE hProcess = INVALID_HANDLE_VALUE;
  CHAR szLine[MAX_LINE + 1];
  INT i;
  DWORD dwExitCode;
  DWORD dwErr = ERROR_SUCCESS;

  dwErr = CreateReadPipe(&hStatusP, &hStatusC);
  if (dwErr != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  dwErr = CreateWritePipe(&hCommandP, &hCommandC);
  if (dwErr != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  dwErr = gpgez_genkey(pgen->szGpgHome, hCommandC, hStatusC, &hProcess);
  if (dwErr != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }

  CloseHandleClean(&hCommandC);
  CloseHandleClean(&hStatusC);

  if ((dwErr = WriteFileLine2(hCommandP, "Key-Type:", "DSA")) != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  if ((dwErr = WriteFileLine2(hCommandP, "Key-Length:", "1024")) != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  if ((dwErr = WriteFileLine2(hCommandP, "Subkey-Type:", "ELG-E")) != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  if ((dwErr = WriteFileLine2(hCommandP, "Subkey-Length:", "2048")) != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  if ((dwErr = WriteFileLine2(hCommandP, "Name-Real:", pgen->pName)) != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  if ((dwErr = WriteFileLine2(hCommandP, "Name-Email:", pgen->pEmail)) != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  if ((pgen->pComment) && (STRLEN(pgen->pComment) > 0)) {
    if ((dwErr = WriteFileLine2(hCommandP, "Name-Comment:", pgen->pComment)) != ERROR_SUCCESS) {
      DEBUGERROR(0, dwErr);
      goto finish_func;
    }
  }
  if ((dwErr = WriteFileLine2(hCommandP, "Expire-Date:", "0")) != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  if (pgen->pPassphrase) {
    if ((dwErr = WriteFileLine2(hCommandP, "Passphrase:", pgen->pPassphrase)) != ERROR_SUCCESS) {
      DEBUGERROR(0, dwErr);
      goto finish_func;
    }
  }
  if ((dwErr = WriteFileLine2(hCommandP, "%secring", strlist_get_node(pgen->pslSecRing, 0))) != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  if ((dwErr = WriteFileLine2(hCommandP, "%pubring", strlist_get_node(pgen->pslPubRing, 0))) != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  if ((dwErr = WriteFileLine1(hCommandP, "%commit")) != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }

  CloseHandleClean(&hCommandP);
  while (dwErr == ERROR_SUCCESS) {
    LPSTR ps, pe;
    UINT icr = 0;
    dwErr = ReadPipeLine(&hStatusP, szLine, MAX_LINE);
    if ((dwErr != ERROR_SUCCESS) && (dwErr != ERROR_HANDLE_EOF)) {
      DEBUGERROR(0, dwErr);
      break;
    }
    if (!STRNCMPL("[GNUPG:] PROGRESS ", szLine)) {
      ps = szLine;
      for (i = 0; i < 3; i++) {
        ps = STRCHR(ps, ' ');
        ps++;
      }
      ShowCharMessage(pgen->hwndMsg, *ps);
    } else if (!STRNCMPL("[GNUPG:] KEY_CREATED ", szLine)) {
      CHAR szLine2[MAX_LINE + 1];
      ps = szLine;

      for (i = 0; i < 3; i++) {
        ps = STRCHR(ps, ' ');
        ps++;
      }
      for (pe = ps; (*pe >= '0' && *pe <= '9') || (*pe >= 'a' && *pe <= 'z') || (*pe >= 'A' && *pe <= 'Z'); pe++)
        ;
      pgen->pFpr = (LPSTR)MALLOC(pe - ps + 1);
      if (!(pgen->pFpr)) {
        dwErr = ERROR_OUTOFMEMORY;
        DEBUGERROR(0, dwErr);
        goto finish_func;
      }
      STRNCPY(pgen->pFpr, ps, pe - ps);
      *(pgen->pFpr + (pe - ps)) = '\0';
      LoadString(g_hInstance, IDS_GPG_KEYCREATED, szLine2, MAX_LINE - 1);
      ShowLineMessage(pgen->hwndMsg, szLine2);
      ShowLineMessage(pgen->hwndMsg, pgen->pFpr);
    } else {
      InterpretAndSendGpgMsg(pgen->hwndMsg, szLine, INVALID_HANDLE_VALUE, NULL);
    }

    if (dwErr == ERROR_HANDLE_EOF) {
      dwErr = ERROR_SUCCESS;
      break;
    }
  }
  WaitForSingleObject(hProcess, INFINITE);
  if (!GetExitCodeProcess(hProcess, &dwExitCode)) {
    dwErr = GetLastError();
    goto finish_func;
  }
  CloseHandleClean(&hProcess);
finish_func:
  if (hProcess != INVALID_HANDLE_VALUE) {
    TerminateProcess(hProcess, 0);
  }
  CloseHandleClean(&hProcess);
  CloseHandleClean(&hCommandC);
  CloseHandleClean(&hCommandP);
  CloseHandleClean(&hStatusC);
  CloseHandleClean(&hStatusP);
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
genkey_exportPubKey(LPCSTR pConfDir, PSTRLIST pPubring, LPSTR pFpr, LPSTR pName) {
  OPENFILENAME ofn;
  CHAR szFileTitle[MAX_PATH + 1] = "";
  CHAR szFileName[MAX_PATH + 1] = "";
  HANDLE hFile = INVALID_HANDLE_VALUE;
  HANDLE hProcess = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;
  SECURITY_ATTRIBUTES sa;

  STRCPY(szFileTitle, pName);
  if (STRLEN(pName) < MAX_PATH - 5) {
    STRCAT(szFileTitle, ".gpg");
  }
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = NULL;
  ofn.lpstrFilter = "gpg(*.gpg)\0*.gpg\0All files(*.*)\0*.*\0\0";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT;
  ofn.lpstrDefExt = "gpg";
  if (GetSaveFileName(&ofn)) {
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.lpSecurityDescriptor = NULL;
    sa.bInheritHandle = TRUE;
    hFile = CreateFile(ofn.lpstrFile, GENERIC_WRITE, 0, &sa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE) {
      dwErr = GetLastError();
      DEBUGERROR(0, dwErr);
      goto finish_func;
    }
    dwErr = gpgez_export(pConfDir, pPubring, hFile, pFpr, &hProcess);
    if (dwErr != ERROR_SUCCESS) {
      DEBUGERROR(0, dwErr);
      goto finish_func;
    }
    WaitForSingleObject(hProcess, INFINITE);
    CloseHandleClean(&hProcess);
    CloseHandleClean(&hFile);
  } else {
    /* G[`FbN */
  }
finish_func:
  if (hProcess != INVALID_HANDLE_VALUE) {
    TerminateProcess(hProcess, 0);
  }
  CloseHandleClean(&hProcess);
  CloseHandleClean(&hFile);
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
genkey_dwGenerateGPGKeys(HWND hwnd) {
  PPARAM_GENKEY pgen;
  CHAR szName[MAX_PATH + 1];
  LPSTR pNameUtf8 = NULL;
  CHAR szEmail[MAX_PATH + 1];
  CHAR szComment[MAX_PATH + 1];
  LPSTR pCommentUtf8 = NULL;
  DWORD dwErr = ERROR_SUCCESS;
  HCURSOR hOldCursor;

  pgen = (PPARAM_GENKEY)GetWindowLong(hwnd, GWL_USERDATA);

  pgen->pFpr = NULL;

  Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_NAME), szName, MAX_PATH);
  strtrim(szName);

  Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_EMAIL), szEmail, MAX_PATH);
  strtrim(szEmail);

  Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_COMMENT), szComment, MAX_PATH);
  strtrim(szComment);

  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_NAME), FALSE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_EMAIL), FALSE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_COMMENT), FALSE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDOK), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), FALSE);
  Edit_SetText(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), "");

  hOldCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));

  /* szName ̏`FbN */
  if ((dwErr = validateName(szName)) != ERROR_SUCCESS) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_INVALIDUSERNAME, m_buff, MAX_PATH) == 0) {
      STRCPY(m_buff, "Invalid User Name");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK);
    goto finish_func;
  }

  /* szEmail ̏`FbN */
  if ((dwErr = validateEmail(szEmail)) != ERROR_SUCCESS) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_INVALIDUSEREMAIL, m_buff, MAX_PATH) == 0) {
      STRCPY(m_buff, "Invalid User Email Address Format");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK);
    goto finish_func;
  }

  /* szComment ̏`FbN */
  if ((dwErr = validateComment(szComment)) != ERROR_SUCCESS) {
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_ERROR_INVALIDUSERCOMMENT, m_buff, MAX_PATH) == 0) {
      STRCPY(m_buff, "Invalid User Comment");
    }
    MessageBox(hwnd, m_buff, "Error", MB_OK);
    goto finish_func;
  }

  if (!(pNameUtf8 = sjis2utf8(szName))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  pgen->pName = pNameUtf8;
  pgen->pEmail = szEmail;
  if (!(pCommentUtf8 = sjis2utf8(szComment))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  pgen->pComment = pCommentUtf8;
  pgen->pPassphrase = NULL;
  pgen->hwndMsg = GetDlgItem(hwnd, IDC_EDIT_MESSAGE);
  dwErr = genkey_dwGenerateGPGKeys2(pgen);
  /* genkey_dwGenerateGPGKeys2()ŃG[\͊Ă̂ŁAł̓G[\͂Ȃ */

  SetCursor(hOldCursor);

  if (dwErr == ERROR_SUCCESS) {
    if (pgen->pFpr) {
      int iRet;
      CHAR szLine[MAX_LINE + 1], szTitle[MAX_LINE + 1];
      LoadString(g_hInstance, IDS_PROMPT_KEYGENSUCCEEDED, szLine, MAX_LINE);
      LoadString(g_hInstance, IDS_TITLE_KEYGENSUCCEEDED, szTitle, MAX_LINE);
      iRet = MessageBox(pgen->hwnd, szLine, szTitle, MB_YESNO | MB_ICONQUESTION);
      if (iRet == IDYES) {
        dwErr = genkey_exportPubKey(pgen->szGpgHome, pgen->pslPubRing, pgen->pFpr, szName);
        /* genkey_exportPubKey()ŃG[\͊Ă̂ŁAł̓G[\͂Ȃ */
      }
    }
  }
finish_func:
  if (pNameUtf8) FREE(pNameUtf8);
  if (pCommentUtf8) FREE(pCommentUtf8);
  if (pgen->pFpr) FREE(pgen->pFpr);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_NAME), TRUE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_EMAIL), TRUE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_COMMENT), TRUE);
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), FALSE);
  Button_Enable(GetDlgItem(hwnd, IDOK), TRUE);
  Button_Enable(GetDlgItem(hwnd, IDCANCEL), TRUE);
  return dwErr;
}

/*------------------------------------------------------------*/
static BOOL
genkey_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  PPARAM_GENKEY pgen = (PPARAM_GENKEY)lParam;

  pgen->hwnd = hwnd;
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_NAME), MAX_PATH);
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_EMAIL), MAX_PATH);
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_COMMENT), MAX_PATH);
  Edit_SetText(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), "");
  Edit_Enable(GetDlgItem(hwnd, IDC_EDIT_MESSAGE), FALSE);
  SetWindowLong(hwnd, GWL_USERDATA, (LPARAM)pgen);
  return TRUE;
}

/*------------------------------------------------------------*/
static VOID
genkey_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  DWORD dwErr;

  switch(id) {
  case IDOK:
    dwErr = genkey_dwGenerateGPGKeys(hwnd);
    if (dwErr != ERROR_SUCCESS) {
    } else {
      EndDialog(hwnd, (INT)ERROR_SUCCESS);
    }
    break;
  case IDCANCEL:
    EndDialog(hwnd, (INT)ERROR_CANCELLED);
    break;
  }
}

/*------------------------------------------------------------*/
static BOOL CALLBACK
procDlgGenkey(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_MSG(hDlg, WM_INITDIALOG, genkey_Cls_OnInitDialog);
    HANDLE_MSG(hDlg, WM_COMMAND, genkey_Cls_OnCommand);
  default:
    return FALSE;
  }
  return TRUE;
}

/*------------------------------------------------------------*/
DWORD
generateGpgKey(HWND hwnd, LPCSTR szGpgHome, PSTRLIST pslSecRing, PSTRLIST pslPubRing) {
  DWORD dwErr = ERROR_SUCCESS;
  PARAM_GENKEY gen;

  gen.szGpgHome = szGpgHome;
  gen.pslSecRing = pslSecRing;
  gen.pslPubRing = pslPubRing;
  dwErr = (DWORD)DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_GENKEY), hwnd, procDlgGenkey, (LPARAM)&gen);
  return dwErr;
}
/*------------------------------------------------------------*/

