/**
 * gpgez.h
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#ifndef GPGEZ_H
#define GPGEZ_H
#include <windows.h>
#include "util.h"

#define GPGBINDIR "bin"
#define GPGCOMMAND "gpg.exe"
#define GPGCOMMANDLINE " --no-random-seed-file --cipher-algo=aes256 "
#define TARCOMMAND "bsdtar.exe"

#define MAX_FPR 1024

enum {
  GNUPG_ALREADY_SIGNED = 1,
  GNUPG_ATTRIBUTE,
  GNUPG_BACKUP_KEY_CREATED,
  GNUPG_BAD_PASSPHRASE,
  GNUPG_BADARMOR,
  GNUPG_BADSIG,
  GNUPG_BEGIN_DECRYPTION,
  GNUPG_BEGIN_ENCRYPTION,
  GNUPG_BEGIN_STREAM,
  GNUPG_CARDCTRL,
  GNUPG_DECRYPTION_FAILED,
  GNUPG_DECRYPTION_OKAY,
  GNUPG_DELETE_PROBLEM,
  GNUPG_ENC_TO,
  GNUPG_END_DECRYPTION,
  GNUPG_END_ENCRYPTION,
  GNUPG_END_STREAM,
  GNUPG_ERROR,
  GNUPG_ERRSIG,
  GNUPG_EXPKEYSIG,
  GNUPG_EXPSIG,
  GNUPG_FILE_DONE,
  GNUPG_FILE_START,
  GNUPG_GET_BOOL,
  GNUPG_GET_HIDDEN,
  GNUPG_GET_LINE,
  GNUPG_GOOD_PASSPHRASE,
  GNUPG_GOODSIG,
  GNUPG_GOT_IT,
  GNUPG_IMPORT_CHECK,
  GNUPG_IMPORT_OK,
  GNUPG_IMPORT_PROBLEM,
  GNUPG_IMPORT_RES,
  GNUPG_IMPORTED,
  GNUPG_INV_RECP,
  GNUPG_KEY_CREATED,
  GNUPG_KEY_NOT_CREATED,
  GNUPG_KEYEXPIRED,
  GNUPG_KEYREVOKED,
  GNUPG_MISSING_PASSPHRASE,
  GNUPG_NEED_PASSPHRASE,
  GNUPG_NEED_PASSPHRASE_PIN,
  GNUPG_NEED_PASSPHRASE_SYM,
  GNUPG_NEWSIG,
  GNUPG_NO_PUBKEY,
  GNUPG_NO_RECP,
  GNUPG_NO_SECKEY,
  GNUPG_NODATA,
  GNUPG_NOTATION_DATA,
  GNUPG_NOTATION_NAME,
  GNUPG_PLAINTEXT,
  GNUPG_PLAINTEXT_LENGTH,
  GNUPG_POLICY_URL,
  GNUPG_PROGRESS,
  GNUPG_REVKEYSIG,
  GNUPG_RSA_OR_IDEA,
  GNUPG_SC_OP_FAILURE,
  GNUPG_SC_OP_SUCCESS,
  GNUPG_SESSION_KEY,
  GNUPG_SHM_GET,
  GNUPG_SHM_GET_BOOL,
  GNUPG_SHM_GET_HIDDEN,
  GNUPG_SHM_INFO,
  GNUPG_SIG_CREATED,
  GNUPG_SIG_ID,
  GNUPG_SIG_SUBPACKET,
  GNUPG_SIGEXPIRED,
  GNUPG_TRUNCATED,
  GNUPG_TRUST_FULLY,
  GNUPG_TRUST_MARGINAL,
  GNUPG_TRUST_NEVER,
  GNUPG_TRUST_ULTIMATE,
  GNUPG_TRUST_UNDEFINED,
  GNUPG_UNEXPECTED,
  GNUPG_USERID_HINT,
  GNUPG_VALIDSIG,
  GNUPG_ZZZZZZZZ
};

struct _GPGMSGS {
  LPSTR smsg;
  UINT nmsg;
};

typedef struct _TAR_PARAM {
  LPCSTR basedir;
  PSTRLIST files;
  HANDLE hStdErr;
} TAR_PARAM, *PTAR_PARAM;

typedef struct _UNTAR_PARAM {
  LPCSTR basedir;
  BOOL fOverwrite;
  /* DWORD dwOverwrite;  0:query 1:always overwrite 2:always skip */
  HANDLE hStdErr;
} UNTAR_PARAM, *PUNTAR_PARAM;

typedef struct _GPGIMPORT_PARAM {
  LPCSTR pszGpgHome;
  PSTRLIST pSecRing;
  PSTRLIST pPubRing;
  HANDLE hCommand;
  HANDLE hStatus;
} GPGIMPORT_PARAM, *PGPGIMPORT_PARAM;

typedef struct _GPGDELETE_PARAM {
  LPCSTR pszGpgHome;
  PSTRLIST pSecRing;
  PSTRLIST pPubRing;
  LPCSTR pFpr;
  HANDLE hCommand;
  HANDLE hStatus;
} GPGDELETE_PARAM, *PGPGDELETE_PARAM;

typedef struct _GPGDECRYPT_PARAM {
  LPCSTR pszGpgHome;
  PSTRLIST pSecRing;
  PSTRLIST pPubRing;
  BOOL fVerify;
  HANDLE hCommand;
  HANDLE hStatus;
  LPSTR pPassphrase;
  LPSTR pPassphraseSym;
  LPSTR pLine2;
} GPGDECRYPT_PARAM, *PGPGDECRYPT_PARAM;

typedef struct _GPGENCRYPT_PARAM {
  LPCSTR pszGpgHome;
  PSTRLIST pSecRing;
  PSTRLIST pPubRing;
  BOOL fNoSign;
  PSTRLIST pFpr;
  HANDLE hCommand;
  HANDLE hStatus;
  LPSTR pPassphrase;
  LPSTR pPassphraseSym;
  LPSTR pLine2;
} GPGENCRYPT_PARAM, *PGPGENCRYPT_PARAM;

typedef struct _CHAINEDPROCESS_NODE {
  struct _CHAINEDPROCESS_NODE *next;
  HANDLE hProcess;
  DWORD dwExitCode; /* vZXpExitCode */
  PHANDLE phObserve; /* Ďpnh */
  DWORD (*dwStartUp)(LPVOID buff);
  DWORD (*dwProcObserved)(LPVOID buff, HWND hwndmsg);
  DWORD (*dwFinishUp)(LPVOID buff);
  HANDLE hRead;
  HANDLE hWrite;
  LPVOID buff;
} CHAINEDPROCESS_NODE, *PCHAINEDPROCESS_NODE;

/*------------------------------------------------------------*/
/* For Archive */
/*------------------------------------------------------------*/
DWORD initializeTarParam(PTAR_PARAM ptparam, LPCSTR pBaseDirectory, PSTRLIST pFiles);
DWORD startupTar(LPVOID buff);
DWORD procObservedTar(LPVOID buff, HWND hwndmsg);
DWORD finishupTar(LPVOID buff);

/*------------------------------------------------------------*/
/* For UnArchive */
/*------------------------------------------------------------*/
DWORD initializeUntarParam(PUNTAR_PARAM puparam, LPCSTR pBaseDirectory, BOOL fOverwrite);
DWORD startupUntar(LPVOID buff);
DWORD procObservedUntar(LPVOID buff, HWND hwndmsg);
DWORD finishupUntar(LPVOID buff);

/*------------------------------------------------------------*/
/* For Import Keys */
/*------------------------------------------------------------*/
DWORD initializeGpgImportParam(PGPGIMPORT_PARAM pgparam, LPCSTR pConfDir, PSTRLIST pSecKey, PSTRLIST pPubKey);
DWORD startupGpgImport(LPVOID buff);
DWORD procObservedGpgImport(LPVOID buff, HWND hwndmsg);
DWORD finishupGpgImport(LPVOID buff);

/*------------------------------------------------------------*/
/* For Delete Keys */
/*------------------------------------------------------------*/
DWORD initializeGpgDeleteParam(PGPGDELETE_PARAM pgparam, LPCSTR pConfDir, PSTRLIST pSecKey, PSTRLIST pPubKey, LPCSTR pFpr);
DWORD startupGpgDelete(LPVOID buff);
DWORD procObservedGpgDelete(LPVOID buff, HWND hwndmsg);
DWORD finishupGpgDelete(LPVOID buff);

/*------------------------------------------------------------*/
/* For Decrypt */
/*------------------------------------------------------------*/
DWORD initializeGpgDecryptParam(PGPGDECRYPT_PARAM pgparam, LPCSTR pConfDir, PSTRLIST pSecKey, PSTRLIST pPubKey, BOOL fVerify, LPSTR pPassphrase, LPSTR pPassphraseSym);
DWORD startupGpgDecrypt(LPVOID buff);
DWORD procObservedGpgDecrypt(LPVOID buff, HWND hwndmsg);
DWORD finishupGpgDecrypt(LPVOID buff);

/*------------------------------------------------------------*/
/* For Encrypt */
/*------------------------------------------------------------*/
DWORD initializeGpgEncryptParam(PGPGENCRYPT_PARAM pgparam, LPCSTR pConfDir, PSTRLIST pSecKey, PSTRLIST pPubKey, PSTRLIST pRecipients, LPSTR pPasphrase, LPSTR pPassphraseSym);
DWORD startupGpgEncrypt(LPVOID buff);
DWORD procObservedGpgEncrypt(LPVOID buff, HWND hwndmsg);
DWORD finishupGpgEncrypt(LPVOID buff);

/*------------------------------------------------------------*/

/* GPG ֐ */
DWORD gpgez_genkey(PCSTR pszGpgHome, HANDLE hStdin, HANDLE hStatus, PHANDLE phProcess);
DWORD gpgez_export(PCSTR pszGpgHome, PSTRLIST pPubring, HANDLE hStdout, PCSTR pFpr, PHANDLE phProcess);
DWORD gpgez_import(PCSTR pszGpgHome, PSTRLIST pSecring, PSTRLIST pPubring, HANDLE hStdin, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess);
DWORD gpgez_delete(PCSTR pszGpgHome, PSTRLIST pPubring, PCSTR pFpr, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess);
DWORD gpgez_listkeys(PCSTR pszGpgHome, PSTRLIST pRing, BOOL fSecret, HANDLE hStdout, PHANDLE phProcess);
DWORD gpgez_decrypt(PCSTR pszGpgHome, PSTRLIST pSecring, PSTRLIST pPubring, BOOL fVerify, HANDLE hStdin, HANDLE hStdout, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess);
DWORD gpgez_encrypt(PCSTR pszGpgHome, PSTRLIST pSecring, PSTRLIST pPubring, BOOL fNoSign, PSTRLIST pRecipients, HANDLE hStdin, HANDLE hStdout, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess);
DWORD gpgez_encrypt_sym(PCSTR pszGpgHome, PSTRLIST pSecring, PSTRLIST pPubring, BOOL fNoSign, HANDLE hStdin, HANDLE hStdout, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess);
DWORD gpgez_sign(PCSTR pszGpgHome, PSTRLIST pSecring, PSTRLIST pPubring, HANDLE hStdin, HANDLE hStdout, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess);
DWORD gpgez_verify(PCSTR pszGpgHome, PSTRLIST pSecring, PSTRLIST pPubring, PCSTR pszSigFile, HANDLE hStdin, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess);

/* tar ֐ */
DWORD gpgez_tar(PCSTR basedir, PSTRLIST files, HANDLE hStdOut, HANDLE hStdErr, PHANDLE phProcess);
DWORD gpgez_untar(PCSTR basedir, HANDLE hStdIn, HANDLE hStdErr, BOOL fOverwrite, PHANDLE phProcess);


DWORD strlist_add_node_dq(PBUFFLIST plist, PCSTR pStr);
DWORD strlist_add_node_ks(PBUFFLIST plist, PCSTR pKey, PCSTR pVal);
DWORD strlist_add_node_ksdq(PBUFFLIST plist, PCSTR pKey, PCSTR pVal);
DWORD strlist_add_node_kd(PBUFFLIST plist, PCSTR pKey, INT i);
DWORD setup_si(STARTUPINFO *psi, PHANDLE phIn, PHANDLE phOut, PHANDLE phNull);

DWORD interpretGpgMsg(LPCSTR msg);


#endif
