/**
 * operateFiles.c
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <windows.h>
#include <windowsx.h>
#include <shlwapi.h>
#include <ShlObj.h>

#include "util.h"
#include "gpgez.h"
#include "JanusHS.h"
#include "resource.h"

extern HINSTANCE g_hInstance;
extern HWND g_hwnd;
extern JANUS_GV gVar;

/*------------------------------------------------------------*/
DWORD
zgpencryptfiles(HWND hwnd, LPSTR pBaseDirectory, PSTRLIST pFiles, LPCSTR pZfile, PSTRLIST pRecipients, LPSTR pPassphraseSym) {
  PCHAINEDPROCESS_NODE phead = NULL, pp = NULL, pp1, ppt;
  TAR_PARAM tparam;
  GPGENCRYPT_PARAM gparam;
  SECURITY_ATTRIBUTES sa;
  HANDLE hRead = INVALID_HANDLE_VALUE, hWrite = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pp1 = (PCHAINEDPROCESS_NODE)MALLOC(sizeof(CHAINEDPROCESS_NODE)))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  initializeTarParam(&tparam, pBaseDirectory, pFiles);
  pp1->buff = (PVOID)&tparam;
  pp1->dwStartUp = startupTar;
  pp1->dwProcObserved = procObservedTar;
  pp1->dwFinishUp = finishupTar;
  pp1->phObserve = NULL;
  pp1->next = NULL;
  phead = pp1;

  if (!(pp1 = (PCHAINEDPROCESS_NODE)MALLOC(sizeof(CHAINEDPROCESS_NODE)))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  initializeGpgEncryptParam(&gparam, gVar.pConfDir, gVar.secKey, gVar.pubKey, pRecipients, NULL, pPassphraseSym);
  pp1->buff = (PVOID)&gparam;
  pp1->dwStartUp = startupGpgEncrypt;
  pp1->dwProcObserved = procObservedGpgEncrypt;
  pp1->dwFinishUp = finishupGpgEncrypt;
  pp1->phObserve = NULL;
  pp1->next = NULL;
  phead->next = pp1;

  /* 1Ԗڂ̃vZXp̕W͂ݒ */
  hRead = INVALID_HANDLE_VALUE;

  /* Ō̃vZX̕Wo͂ݒ */
  ZeroMemory(&sa, sizeof(sa));
  sa.nLength = sizeof(sa);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;
  if ((hWrite = CreateFile(pZfile, GENERIC_WRITE, 0, &sa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) {
    dwErr = GetLastError();
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }

  dwErr = raiseupChainedProcess(hwnd, phead, &hRead, &hWrite);
  /* raiseupChainedProcess()ŃG[\͊Ă̂ŁAł͕\Ȃ */

finish_func:
  for (pp1 = phead; pp1; pp1 = pp1->next) {
    ppt = pp1->next;
    FREE(pp1);
    pp1 = ppt;
  }
  if (hRead != INVALID_HANDLE_VALUE) CloseHandleClean(&hRead);
  if (hWrite != INVALID_HANDLE_VALUE) CloseHandleClean(&hWrite);
  return dwErr;
}

/*------------------------------------------------------------*/
DWORD
zgpencrypt(HWND hwnd, PJANUS_FILELIST pjfl, BOOL fPublic) {
  PSTRLIST recipients = NULL;
  DWORD dwErr = ERROR_SUCCESS;
  CHAR szFileName[MAX_PATH + 1] = "";
  OPENFILENAME ofn;
  DWORD dwEncType = ENCTYPE_NONE;

  if (bufflist_getcount(pjfl->files) < 1) {
    dwErr = ERROR_INVALID_DATA;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }

  /* Recipients w */
  if (!(recipients = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  if (fPublic) {
    dwErr = selectGpgKey(hwnd, gVar.pConfDir, gVar.secKey, gVar.pubKey, recipients, &dwEncType);
    if (dwErr != ERROR_SUCCESS) {
      DEBUGERROR(0, dwErr);
      goto finish_func;
    }
  } else {
    CHAR buff[MAX_PATH + 1] = "";
    CHAR m_buff[MAX_PATH + 1] = "";

    if (LoadString(g_hInstance, IDS_NEEDPASSPHRASESYM, m_buff, MAX_PATH) == 0) {
      STRCPY(m_buff, "Passphrase");
    }
    if ((dwErr = InputBoxEx(hwnd, m_buff, "Symmetric Encryption", buff, MAX_PATH, INPUTBOX_SECRET | INPUTBOX_CONFIRM)) != ERROR_SUCCESS) {
      /* G[ƂĂERROR_CANCELLED炢Ȃ̂ŁAĕ\͂Ȃ */
      goto finish_func;
    }
    if (STRLEN(buff) < 4) {
      if (LoadString(g_hInstance, IDS_ERROR_PASSPHRASETOOSHORT, m_buff, MAX_PATH) == 0) {
        STRCPY(m_buff, "Passphrase too short");
      }
      MessageBox(g_hwnd, m_buff, "Error", MB_OK);
      goto finish_func;
    }
    if ((dwErr = strlist_add_node(recipients, buff)) != ERROR_SUCCESS) {
      DEBUGERROR(0, dwErr);
      goto finish_func;
    }
    dwEncType = ENCTYPE_SYMMETRIC; /* Symmetric Encryption */
  }

  /* ot@Cw */
  ZeroMemory(&ofn, sizeof(OPENFILENAME));
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hwnd;
  ofn.lpstrFilter = "gpg(*.gpg)\0*.gpg\0All files(*.*)\0*.*\0\0";
  ofn.lpstrFile = szFileName;
  ofn.lpstrFileTitle = NULL;
  ofn.nMaxFile = MAX_PATH;
  ofn.nMaxFileTitle = MAX_PATH;
  ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT;
  ofn.lpstrDefExt = "gpg";
  if (GetSaveFileName(&ofn)) {
    if (dwEncType == ENCTYPE_PUBLIC) {
      dwErr = zgpencryptfiles(hwnd, pjfl->basedir, pjfl->files, szFileName, recipients, NULL);
      if (dwErr != ERROR_SUCCESS) {
        DEBUGERROR(0, dwErr);
        goto finish_func;
      }
    } else if (dwEncType == ENCTYPE_SYMMETRIC) {
      dwErr = zgpencryptfiles(hwnd, pjfl->basedir, pjfl->files, szFileName, NULL, strlist_get_node(recipients, 0));
      if (dwErr != ERROR_SUCCESS) {
        if (dwErr != ERROR_CANCELLED) {
          DEBUGERROR(0, dwErr);
          goto finish_func;
        }
      }
    }
  } else {
    /* G[`FbN */
  }
finish_func:
  if (recipients) strlist_release(recipients);
  return dwErr;
}

/*------------------------------------------------------------*/
DWORD
zgpdecryptfiles(HWND hwnd, LPCSTR pFilename, LPCSTR pBaseDirectory, BOOL fOverwrite, BOOL fPublic) {
  PCHAINEDPROCESS_NODE phead = NULL, pp = NULL, pp1, ppt;
  HANDLE hRead = INVALID_HANDLE_VALUE, hWrite = INVALID_HANDLE_VALUE;
  UNTAR_PARAM uparam;
  GPGDECRYPT_PARAM gparam;
  SECURITY_ATTRIBUTES sa;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pp1 = (PCHAINEDPROCESS_NODE)MALLOC(sizeof(CHAINEDPROCESS_NODE)))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  initializeGpgDecryptParam(&gparam, gVar.pConfDir, gVar.secKey, gVar.pubKey, fPublic ? TRUE : FALSE, NULL, NULL);
  pp1->buff = (PVOID)&gparam;
  pp1->dwStartUp = startupGpgDecrypt;
  pp1->dwProcObserved = procObservedGpgDecrypt;
  pp1->dwFinishUp = finishupGpgDecrypt;
  pp1->phObserve = NULL;
  pp1->next = NULL;
  phead = pp1;

  if (!(pp1 = (PCHAINEDPROCESS_NODE)MALLOC(sizeof(CHAINEDPROCESS_NODE)))) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  initializeUntarParam(&uparam, pBaseDirectory, fOverwrite);
  pp1->buff = (PVOID)&uparam;
  pp1->dwStartUp = startupUntar;
  pp1->dwProcObserved = procObservedUntar;
  pp1->dwFinishUp = finishupUntar;
  pp1->phObserve = NULL;
  pp1->next = NULL;
  phead->next = pp1;

  ZeroMemory(&sa, sizeof(sa));
  sa.nLength = sizeof(sa);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;
  if ((hRead = CreateFile(pFilename, GENERIC_READ, 0, &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) {
    dwErr = GetLastError();
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  hWrite = INVALID_HANDLE_VALUE;

  dwErr = raiseupChainedProcess(hwnd, phead, &hRead, &hWrite);
  /* raiseupChainedProcess()ŃG[\͊Ă̂ŁAł͕\Ȃ */

finish_func:
  for (pp1 = phead; pp1; pp1 = pp1->next) {
    ppt = pp1->next;
    FREE(pp1);
    pp1 = ppt;
  }
  if (hRead != INVALID_HANDLE_VALUE) CloseHandleClean(&hRead);
  if (hWrite != INVALID_HANDLE_VALUE) CloseHandleClean(&hWrite);
  return dwErr;
}

/*------------------------------------------------------------*/
static WNDPROC procOld;
#define IDC_OVERWRITE 1001

typedef struct _OVERWRITECHECK {
  HWND hOverwrite;
  BOOL fOverwrite;
} OVERWRITECHECK, *POVERWRITECHECK;

/*------------------------------------------------------------*/
static int CALLBACK
procSelectUntarFolderUser(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {

  if ((uMsg == WM_COMMAND) && (LOWORD(wParam) == IDC_OVERWRITE)) {
    POVERWRITECHECK poverwrite;

    poverwrite = (POVERWRITECHECK)GetWindowLong((HWND)lParam, GWL_USERDATA);
    if (Button_GetCheck(poverwrite->hOverwrite)) {
      poverwrite->fOverwrite = FALSE;
    } else {
      poverwrite->fOverwrite = TRUE;
    }
    Button_SetCheck(poverwrite->hOverwrite, poverwrite->fOverwrite);
  } else if (uMsg == WM_DESTROY) {
    SetWindowLong(hwnd, GWL_WNDPROC, (LONG)procOld);
  }
  return(CallWindowProc(procOld, hwnd, uMsg, wParam, lParam ) );
}

/*------------------------------------------------------------*/
static int CALLBACK
procSelectUntarFolder(HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData) {
  POVERWRITECHECK poverwrite = (POVERWRITECHECK)lpData;

  if(uMsg == BFFM_INITIALIZED){
    CHAR m_buff[MAX_PATH + 1];

    if (LoadString(g_hInstance, IDS_CONFIRM_OVERWRITEFILES, m_buff, MAX_PATH) == 0) {
      STRCPY(m_buff, "Overwrite Files");
    }
    poverwrite->hOverwrite = CreateWindow("BUTTON", m_buff, WS_CHILD | WS_VISIBLE | BS_CHECKBOX, 24, 2, 120, 40, hwnd, (HMENU)IDC_OVERWRITE, g_hInstance, NULL);
    poverwrite->fOverwrite = TRUE;
    Button_SetCheck(poverwrite->hOverwrite, poverwrite);
    SendMessage(poverwrite->hOverwrite, WM_SETFONT, (WPARAM)GetStockObject(DEFAULT_GUI_FONT), FALSE);
    procOld = (WNDPROC)SetWindowLong(hwnd, GWL_WNDPROC, (LONG)procSelectUntarFolderUser);
    SetWindowLong(poverwrite->hOverwrite, GWL_USERDATA, (LONG)poverwrite);
  }
  return 0;
}
/*------------------------------------------------------------*/
DWORD
zgpdecrypt(HWND hwnd, LPCSTR pszFilename, BOOL fPublic) {
  DWORD dwErr = ERROR_SUCCESS;
  OVERWRITECHECK overwrite;
  CHAR szDir[MAX_PATH + 1];
  LPITEMIDLIST lpid;
  BROWSEINFO bi;
  LPMALLOC pMalloc = NULL;

  if (E_FAIL == SHGetMalloc(&pMalloc)) {
    dwErr = ERROR_OUTOFMEMORY;
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  /* otH_w */
  ZeroMemory(&bi, sizeof(BROWSEINFO));
  bi.hwndOwner = hwnd;
  bi.lpfn = procSelectUntarFolder;
  bi.lpszTitle = "";
  bi.ulFlags = BIF_RETURNONLYFSDIRS | BIF_NEWDIALOGSTYLE;
  bi.lParam = (LPARAM)&overwrite;
  if (lpid = SHBrowseForFolder(&bi)) {
    SHGetPathFromIDList(lpid, szDir);
    /*
    if (szDir[STRLEN(szDir) - 1] != '\\')
      strcat(szDir, "\\");
     */
    pMalloc->lpVtbl->Free(pMalloc, lpid);
    pMalloc->lpVtbl->Release(pMalloc);
    dwErr = zgpdecryptfiles(hwnd, pszFilename, szDir, overwrite.fOverwrite, fPublic);
    if (dwErr != ERROR_SUCCESS) goto finish_func;
  } else {
    dwErr = ERROR_CANCELLED;
  }
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/

