/**
 * rungpg.c
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <windows.h>
#include "util.h"
#include "gpgez.h"

struct _GPGMSGS GPGMSGS[] = {
  {"ALREADY_SIGNED", GNUPG_ALREADY_SIGNED},
  {"ATTRIBUTE", GNUPG_ATTRIBUTE},
  {"BACKUP_KEY_CREATED", GNUPG_BACKUP_KEY_CREATED},
  {"BAD_PASSPHRASE", GNUPG_BACKUP_KEY_CREATED},
  {"BADARMOR", GNUPG_BADARMOR},
  {"BADSIG", GNUPG_BADSIG},
  {"BEGIN_DECRYPTION", GNUPG_BEGIN_DECRYPTION},
  {"BEGIN_ENCRYPTION", GNUPG_BEGIN_ENCRYPTION},
  {"BEGIN_STREAM", GNUPG_BEGIN_STREAM},
  {"CARDCTRL", GNUPG_CARDCTRL},
  {"DECRYPTION_FAILED", GNUPG_DECRYPTION_FAILED},
  {"DECRYPTION_OKAY", GNUPG_DECRYPTION_OKAY},
  {"DELETE_PROBLEM", GNUPG_DELETE_PROBLEM},
  {"ENC_TO", GNUPG_ENC_TO},
  {"END_DECRYPTION", GNUPG_END_DECRYPTION},
  {"END_ENCRYPTION", GNUPG_END_ENCRYPTION},
  {"END_STREAM", GNUPG_END_STREAM},
  {"ERROR", GNUPG_ERROR},
  {"ERRSIG", GNUPG_ERRSIG},
  {"EXPKEYSIG", GNUPG_EXPKEYSIG},
  {"EXPSIG", GNUPG_EXPSIG},
  {"FILE_DONE", GNUPG_FILE_DONE},
  {"FILE_START", GNUPG_FILE_START},
  {"GET_BOOL", GNUPG_GET_BOOL},
  {"GET_HIDDEN", GNUPG_GET_HIDDEN},
  {"GET_LINE", GNUPG_GET_LINE},
  {"GOOD_PASSPHRASE", GNUPG_GOOD_PASSPHRASE},
  {"GOODSIG", GNUPG_GOODSIG},
  {"GOT_IT", GNUPG_GOT_IT},
  {"IMPORT_CHECK", GNUPG_IMPORT_CHECK},
  {"IMPORT_OK", GNUPG_IMPORT_OK},
  {"IMPORT_PROBLEM", GNUPG_IMPORT_PROBLEM},
  {"IMPORT_RES", GNUPG_IMPORT_RES},
  {"IMPORTED", GNUPG_IMPORTED},
  {"INV_RECP", GNUPG_INV_RECP},
  {"KEY_CREATED", GNUPG_KEY_CREATED},
  {"KEY_NOT_CREATED", GNUPG_KEY_NOT_CREATED},
  {"KEYEXPIRED", GNUPG_KEYEXPIRED},
  {"KEYREVOKED", GNUPG_KEYREVOKED},
  {"MISSING_PASSPHRASE", GNUPG_MISSING_PASSPHRASE},
  {"NEED_PASSPHRASE", GNUPG_NEED_PASSPHRASE},
  {"NEED_PASSPHRASE_PIN", GNUPG_NEED_PASSPHRASE_PIN},
  {"NEED_PASSPHRASE_SYM", GNUPG_NEED_PASSPHRASE_SYM},
  {"NEWSIG", GNUPG_NEWSIG},
  {"NO_PUBKEY", GNUPG_NO_PUBKEY},
  {"NO_RECP", GNUPG_NO_RECP},
  {"NO_SECKEY", GNUPG_NO_SECKEY},
  {"NODATA", GNUPG_NODATA},
  {"NOTATION_DATA", GNUPG_NOTATION_DATA},
  {"NOTATION_NAME", GNUPG_NOTATION_NAME},
  {"PLAINTEXT", GNUPG_PLAINTEXT},
  {"PLAINTEXT_LENGTH", GNUPG_PLAINTEXT_LENGTH},
  {"POLICY_URL", GNUPG_POLICY_URL},
  {"PROGRESS", GNUPG_PROGRESS},
  {"REVKEYSIG", GNUPG_REVKEYSIG},
  {"RSA_OR_IDEA", GNUPG_RSA_OR_IDEA},
  {"SC_OP_FAILURE", GNUPG_SC_OP_FAILURE},
  {"SC_OP_SUCCESS", GNUPG_SC_OP_SUCCESS},
  {"SESSION_KEY", GNUPG_SESSION_KEY},
  {"SHM_GET", GNUPG_SHM_GET},
  {"SHM_GET_BOOL", GNUPG_SHM_GET_BOOL},
  {"SHM_GET_HIDDEN", GNUPG_SHM_GET_HIDDEN},
  {"SHM_INFO", GNUPG_SHM_INFO},
  {"SIG_CREATED", GNUPG_SIG_CREATED},
  {"SIG_ID", GNUPG_SIG_ID},
  {"SIG_SUBPACKET", GNUPG_SIG_SUBPACKET},
  {"SIGEXPIRED", GNUPG_SIGEXPIRED},
  {"TRUNCATED", GNUPG_TRUNCATED},
  {"TRUST_FULLY", GNUPG_TRUST_FULLY},
  {"TRUST_MARGINAL", GNUPG_TRUST_MARGINAL},
  {"TRUST_NEVER", GNUPG_TRUST_NEVER},
  {"TRUST_ULTIMATE", GNUPG_TRUST_ULTIMATE},
  {"TRUST_UNDEFINED", GNUPG_TRUST_UNDEFINED},
  {"UNEXPECTED", GNUPG_UNEXPECTED},
  {"USERID_HINT", GNUPG_USERID_HINT},
  {"VALIDSIG", GNUPG_VALIDSIG},
  {NULL, 0}
};

#ifdef ERRCHECK
#undef ERRCHECK
#endif
#define ERRCHECK(d, e) if(((d) = (e)) != ERROR_SUCCESS) goto finish_func;

/*------------------------------------------------------------*/
DWORD
interpretGpgMsg(LPCSTR msg) {
  UINT i = 0;
  LPCSTR p;

  if (STRNCMPL("[GNUPG:] ", msg)) return 0;
  while (GPGMSGS[i].smsg) {
    if (!STRNCMPL(GPGMSGS[i].smsg, msg + STRLEN("[GNUPG:] "))) {
      p = msg + STRLEN("[GNUPG:] ") + STRLEN(GPGMSGS[i].smsg);
      if ((*p == '\0') || (*p == ' ') || (*p =='\n')) {
        break;
      }
    }
    i++;
  }
  return GPGMSGS[i].nmsg;
}

/*------------------------------------------------------------*/
DWORD
strlist_add_node_dq(PSTRLIST plist, LPCSTR pStr) {
  LPSTR s = NULL;
  DWORD dwErr = ERROR_SUCCESS;

  s = (LPSTR)MALLOC(STRLEN(pStr) + 3);
  if (!s) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  STRCPY(s, "\"\0");
  STRCAT(s, pStr);
  STRCAT(s, "\"");
  dwErr = strlist_add_node(plist, s);
  FREE(s);
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/
DWORD
strlist_add_node_ks(PSTRLIST plist, LPCSTR pKey, LPCSTR pVal) {
  LPSTR s = NULL;
  DWORD dwErr = ERROR_SUCCESS;

  s = (LPSTR)MALLOC(STRLEN(pKey) + STRLEN(pVal) + 2);
  if (!s) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  STRCPY(s, pKey);
  STRCAT(s, "=");
  STRCAT(s, pVal);
  dwErr = strlist_add_node(plist, s);
  FREE(s);
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/
DWORD
strlist_add_node_ksdq(PSTRLIST plist, LPCSTR pKey, LPCSTR pVal) {
  LPSTR s = NULL;
  DWORD dwErr = ERROR_SUCCESS;

  s = (LPSTR)MALLOC(STRLEN(pKey) + STRLEN(pVal) + 4);
  if (!s) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  STRCPY(s, pKey);
  STRCAT(s, "=\"");
  STRCAT(s, pVal);
  STRCAT(s, "\"");
  dwErr = strlist_add_node(plist, s);
  FREE(s);
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/
DWORD
strlist_add_node_kd(PSTRLIST plist, LPCSTR pKey, INT i) {
  LPSTR s = NULL, s1;
  INT n, ii, n0;
  DWORD dwErr = ERROR_SUCCESS;

  for (n = 0, ii = i; ii; ii /= 10)
    n++;
  s = (LPSTR)MALLOC(STRLEN(pKey) + n + 2);
  if (!s) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  STRCPY(s, pKey);
  STRCAT(s, "=");
  for (s1 = s; *s1; s1++)
    ;
  for (n0 = 0, ii = i; n0 < n; n0++) {
    *(s1 + n - n0 - 1) = '0' + ii % 10;
    ii /= 10;
  }
  *(s1 + n) = '\0';
  dwErr = strlist_add_node(plist, s);
  FREE(s);
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
setup_si(STARTUPINFO *psi, PHANDLE phIn, PHANDLE phOut, PHANDLE phNull) {
  SECURITY_ATTRIBUTES sa;
  DWORD dwErr = ERROR_SUCCESS;

  ZeroMemory(&sa, sizeof(SECURITY_ATTRIBUTES));
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.bInheritHandle = TRUE;

  *phNull = CreateFile ("nul",
                        GENERIC_READ|GENERIC_WRITE,
                        FILE_SHARE_READ|FILE_SHARE_WRITE,
                        &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
  if (*phNull == INVALID_HANDLE_VALUE) {
    dwErr = GetLastError();
    goto finish_func;
  }

  ZeroMemory(psi, sizeof(STARTUPINFO));
  psi->cb = sizeof(STARTUPINFO);
  psi->dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
  psi->wShowWindow = SW_HIDE;
  psi->hStdInput = (phIn == NULL) ? *phNull : *phIn;
  psi->hStdOutput = (phOut == NULL) ? *phNull : *phOut;
  psi->hStdError = *phNull;
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/
static DWORD
rungpg(LPCSTR pCommandLineOptions, STARTUPINFO *psi, PROCESS_INFORMATION *ppi) {
  CHAR szGpgCommandFile[MAX_PATH + 1];
  CHAR szHomeDirectory[MAX_PATH + 1];
  DWORD dwErr = ERROR_SUCCESS;

  GetModuleFileName(NULL, szHomeDirectory, MAX_PATH);
  *(strryen(szHomeDirectory) + 1) = '\0';
  STRCPY(szGpgCommandFile, szHomeDirectory);
  STRCAT(szGpgCommandFile, GPGBINDIR);
  STRCAT(szGpgCommandFile, "\\");
  STRCAT(szGpgCommandFile, GPGCOMMAND);
  if (!CreateProcess(szGpgCommandFile, (LPTSTR)pCommandLineOptions, NULL, NULL, TRUE, 0, NULL, szHomeDirectory, psi, ppi)) {
    dwErr = GetLastError();
    goto finish_func;
  }
  CloseHandleClean(&(ppi->hThread));
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/
/* For Genkey */
/*------------------------------------------------------------*/
static LPSTR
gpgez_genkey_generate_commandline(LPCSTR pszGpgHome, INT iStatusFd) {
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, GPGCOMMANDLINE));
  ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--homedir", pszGpgHome));
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--status-fd", iStatusFd));
  ERRCHECK(dwErr, strlist_add_node(pOption, "--batch --gen-key"));
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  if (pOption) strlist_release(pOption);
  if (dwErr != ERROR_SUCCESS) {
    FREE(pCommandLine);
  }
  return pCommandLine;
}

/*------------------------------------------------------------*/
DWORD
gpgez_genkey(LPCSTR pszGpgHome, HANDLE hStdin, HANDLE hStatus, PHANDLE phProcess) {
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  LPSTR pCommandLine = NULL;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  ERRCHECK(dwErr, setup_si(&si, &hStdin, NULL, &hNull));
  if (!(pCommandLine = gpgez_genkey_generate_commandline(pszGpgHome, (INT)hStatus))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  pi.hProcess = INVALID_HANDLE_VALUE;
  ERRCHECK(dwErr, rungpg(pCommandLine, &si, &pi));
  CloseHandleClean(&hNull);

  *phProcess = pi.hProcess;

finish_func:
  FREE(pCommandLine);

  return dwErr;
}

/*------------------------------------------------------------*/
/* For Export */
/*------------------------------------------------------------*/
static LPSTR
gpgez_export_generate_commandline(LPCSTR pszGpgHome, PSTRLIST pListPubRing, LPCSTR pFpr) {
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  PSTRLIST_ITERATOR pitr;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, GPGCOMMANDLINE));
  ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--homedir", pszGpgHome));
  for (pitr = strlist_itr_gethead(pListPubRing) ; pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--keyring", strlist_itr_getstr(pitr)));
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, "--no-default-keyring --batch --yes --no-default-keyring --export"));
  ERRCHECK(dwErr, strlist_add_node(pOption, pFpr));
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  if (pOption) strlist_release(pOption);
  if (dwErr != ERROR_SUCCESS) {
    FREE(pCommandLine);
  }
  return pCommandLine;
}

/*------------------------------------------------------------*/
DWORD
gpgez_export(LPCSTR pszGpgHome, PSTRLIST pPubring, HANDLE hStdout, LPCSTR pFpr, PHANDLE phProcess) {
  LPSTR pCommandLine = NULL;
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  ERRCHECK(dwErr, setup_si(&si, NULL, &hStdout, &hNull));
  if (!(pCommandLine = gpgez_export_generate_commandline(pszGpgHome, pPubring, pFpr))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  pi.hProcess = INVALID_HANDLE_VALUE;
  ERRCHECK(dwErr, rungpg(pCommandLine, &si, &pi));
  CloseHandleClean(&hNull);

  *phProcess = pi.hProcess;

finish_func:
  FREE(pCommandLine);

  return (dwErr);
}

/*------------------------------------------------------------*/
/* For Import */
/*------------------------------------------------------------*/
static LPSTR
gpgez_import_generate_commandline(LPCSTR pszGpgHome, PSTRLIST pSecring, PSTRLIST pPubring, INT iCommandFd, INT iStatusFd) {
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  PSTRLIST_ITERATOR pitr;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, GPGCOMMANDLINE));
  ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--homedir", pszGpgHome));
  for (pitr = strlist_itr_gethead(pSecring); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--secret-keyring", strlist_itr_getstr(pitr)));
  }
  for (pitr = strlist_itr_gethead(pPubring); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--keyring", strlist_itr_getstr(pitr)));
  }
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--command-fd", iCommandFd));
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--status-fd", iStatusFd));
  ERRCHECK(dwErr, strlist_add_node(pOption, "--no-default-keyring --batch --fast-import"));
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  if (pOption) bufflist_release(pOption);
  if (dwErr != ERROR_SUCCESS) {
    FREE(pCommandLine);
  }
  return pCommandLine;
}

/*------------------------------------------------------------*/
DWORD
gpgez_import(LPCSTR pszGpgHome, PSTRLIST pSecring, PSTRLIST pPubring, HANDLE hStdin, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess) {
  LPSTR pCommandLine = NULL;
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  ERRCHECK(dwErr, setup_si(&si, &hStdin, NULL, &hNull));
  if (!(pCommandLine = gpgez_import_generate_commandline(pszGpgHome, pSecring, pPubring, (INT)hCommand, (INT)hStatus))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  pi.hProcess = INVALID_HANDLE_VALUE;
  ERRCHECK(dwErr, rungpg(pCommandLine, &si, &pi));
  CloseHandleClean(&hNull);

  *phProcess = pi.hProcess;

finish_func:
  FREE(pCommandLine);

  return dwErr;
}

/*------------------------------------------------------------*/
/* For Delete */
/*------------------------------------------------------------*/
static LPSTR
gpgez_delete_generate_commandline(LPCSTR pszGpgHome, PSTRLIST pListPubRing, LPCSTR pFpr, INT iCommandFd, INT iStatusFd) {
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  PSTRLIST_ITERATOR pitr;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, GPGCOMMANDLINE));
  ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--homedir", pszGpgHome));
  for (pitr = strlist_itr_gethead(pListPubRing); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--keyring", strlist_itr_getstr(pitr)));
  }
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--command-fd", iCommandFd));
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--status-fd", iStatusFd));
  ERRCHECK(dwErr, strlist_add_node(pOption, "--batch --yes --no-default-keyring --delete-key"));
  ERRCHECK(dwErr, strlist_add_node(pOption, pFpr));
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  if (pOption) bufflist_release(pOption);
  if (dwErr != ERROR_SUCCESS) {
    FREE(pCommandLine);
  }
  return pCommandLine;
}

/*------------------------------------------------------------*/
DWORD
gpgez_delete(LPCSTR pszGpgHome, PSTRLIST pPubring, LPCSTR pFpr, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess) {
  LPSTR pCommandLine = NULL;
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  ERRCHECK(dwErr, setup_si(&si, NULL, NULL, &hNull));
  if (!(pCommandLine = gpgez_delete_generate_commandline(pszGpgHome, pPubring, pFpr, (INT)hCommand, (INT)hStatus))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  pi.hProcess = INVALID_HANDLE_VALUE;
  ERRCHECK(dwErr, rungpg(pCommandLine, &si, &pi));
  CloseHandleClean(&hNull);

  *phProcess = pi.hProcess;

finish_func:
  FREE(pCommandLine);

  return (dwErr);
}

/*------------------------------------------------------------*/
/* For Listkeys */
/*------------------------------------------------------------*/
static LPSTR
gpgez_listkeys_generate_commandline(LPCSTR pszGpgHome, PSTRLIST pRing, BOOL fSecret) {
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  PSTRLIST_ITERATOR pitr;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, GPGCOMMANDLINE));
  ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--homedir", pszGpgHome));
  if (fSecret) {
    for (pitr = strlist_itr_gethead(pRing); pitr; pitr = strlist_itr_getnext(pitr)) {
      ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--secret-keyring", strlist_itr_getstr(pitr)));
    }
    ERRCHECK(dwErr, strlist_add_node(pOption, "--list-secret-keys"));
  } else {
    for (pitr = strlist_itr_gethead(pRing); pitr; pitr = strlist_itr_getnext(pitr)) {
      ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--keyring", strlist_itr_getstr(pitr)));
    }
    ERRCHECK(dwErr, strlist_add_node(pOption, "--list-keys"));
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, "--batch --no-default-keyring --with-colons --with-fingerprint"));
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  if (pOption) bufflist_release(pOption);
  if (dwErr != ERROR_SUCCESS) {
    FREE(pCommandLine);
  }
  return pCommandLine;
}

/*------------------------------------------------------------*/
DWORD
gpgez_listkeys(LPCSTR pszGpgHome, PSTRLIST pRing, BOOL fSecret, HANDLE hStdout, PHANDLE phProcess) {
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  LPSTR pCommandLine = NULL;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  ERRCHECK(dwErr, setup_si(&si, NULL, &hStdout, &hNull));
  if (!(pCommandLine = gpgez_listkeys_generate_commandline(pszGpgHome, pRing, fSecret))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  pi.hProcess = INVALID_HANDLE_VALUE;
  ERRCHECK(dwErr, rungpg(pCommandLine, &si, &pi));
  CloseHandleClean(&hNull);

  *phProcess = pi.hProcess;

finish_func:
  FREE(pCommandLine);
  return dwErr;
}

/*------------------------------------------------------------*/
/* For Decrypt */
/*------------------------------------------------------------*/
static LPSTR
gpgez_decrypt_generate_commandline(LPCSTR pszGpgHome, PSTRLIST pSecRing, PSTRLIST pPubRing, BOOL fNoVerify, INT iCommandFd, INT iStatusFd) {
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  PSTRLIST_ITERATOR pitr;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, GPGCOMMANDLINE));
  ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--homedir", pszGpgHome));
  for (pitr = strlist_itr_gethead(pSecRing); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--secret-keyring", strlist_itr_getstr(pitr)));
  }
  for (pitr = strlist_itr_gethead(pPubRing); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--keyring", strlist_itr_getstr(pitr)));
  }
  if (fNoVerify) {
    ERRCHECK(dwErr, strlist_add_node(pOption, "--skip-verify"));
  }
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--command-fd", iCommandFd));
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--status-fd", iStatusFd));
  ERRCHECK(dwErr, strlist_add_node(pOption, "--yes --always-trust --decrypt"));
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  if (pOption) strlist_release(pOption);
  if (dwErr != ERROR_SUCCESS) {
    FREE(pCommandLine);
  }
  return pCommandLine;
}

/*------------------------------------------------------------*/
DWORD
gpgez_decrypt(LPCSTR pszGpgHome, PSTRLIST pSecRing, PSTRLIST pPubRing, BOOL fNoVerify, HANDLE hStdin, HANDLE hStdout, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess) {
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  LPSTR pCommandLine = NULL;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  ERRCHECK(dwErr, setup_si(&si, &hStdin, &hStdout, &hNull));
  if (!(pCommandLine = gpgez_decrypt_generate_commandline(pszGpgHome, pSecRing, pPubRing, fNoVerify, (INT)hCommand, (INT)hStatus))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  pi.hProcess = INVALID_HANDLE_VALUE;
  ERRCHECK(dwErr, rungpg(pCommandLine, &si, &pi));
  CloseHandleClean(&hNull);

  *phProcess = pi.hProcess;

finish_func:
  FREE(pCommandLine);
  return dwErr;
}

/*------------------------------------------------------------*/
/* For Encrypt */
/*------------------------------------------------------------*/
static LPSTR
gpgez_encrypt_generate_commandline(LPCSTR pszGpgHome, PSTRLIST pSecRing, PSTRLIST pPubRing, BOOL fNoSign, PSTRLIST pFpr, INT iCommandFd, INT iStatusFd) {
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  PSTRLIST_ITERATOR pitr;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, GPGCOMMANDLINE));
  ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--homedir", pszGpgHome));
  for (pitr = strlist_itr_gethead(pSecRing); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--secret-keyring", strlist_itr_getstr(pitr)));
  }
  for (pitr = strlist_itr_gethead(pPubRing); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--keyring", strlist_itr_getstr(pitr)));
  }
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--command-fd", iCommandFd));
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--status-fd", iStatusFd));
  if (!(fNoSign)) {
    ERRCHECK(dwErr, strlist_add_node(pOption, "--sign"));
  }
  for (pitr = strlist_itr_gethead(pFpr); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ks(pOption, "--recipient", strlist_itr_getstr(pitr)));
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, "--yes --always-trust --compress-level 0 --encrypt"));
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  if (pOption) strlist_release(pOption);
  if (dwErr != ERROR_SUCCESS) {
    FREE(pCommandLine);
  }
  return pCommandLine;
}

/*------------------------------------------------------------*/
DWORD
gpgez_encrypt(LPCSTR pszGpgHome, PSTRLIST pSecRing, PSTRLIST pPubRing, BOOL fNoSign, PSTRLIST pFpr, HANDLE hStdin, HANDLE hStdout, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess) {
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  LPSTR pCommandLine = NULL;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  ERRCHECK(dwErr, setup_si(&si, &hStdin, &hStdout, &hNull));
  if (!(pCommandLine = gpgez_encrypt_generate_commandline(pszGpgHome, pSecRing, pPubRing, fNoSign, pFpr, (INT)hCommand, (INT)hStatus))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  pi.hProcess = INVALID_HANDLE_VALUE;
  ERRCHECK(dwErr, rungpg(pCommandLine, &si, &pi));
  CloseHandleClean(&hNull);

  *phProcess = pi.hProcess;

finish_func:
  FREE(pCommandLine);

  return (dwErr);
}

/*------------------------------------------------------------*/
/* For Symmetric Encrypt */
/*------------------------------------------------------------*/
static LPSTR
gpgez_encrypt_sym_generate_commandline(LPCSTR pszGpgHome, PSTRLIST pSecRing, PSTRLIST pPubRing, BOOL fNoSign, INT iCommandFd, INT iStatusFd) {
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  PSTRLIST_ITERATOR pitr;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, GPGCOMMANDLINE));
  ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--homedir", pszGpgHome));
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--command-fd", iCommandFd));
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--status-fd", iStatusFd));
  if (!(fNoSign)) {
    for (pitr = strlist_itr_gethead(pSecRing); pitr; pitr = strlist_itr_getnext(pitr)) {
      ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--secret-keyring", strlist_itr_getstr(pitr)));
    }
    for (pitr = strlist_itr_gethead(pPubRing); pitr; pitr = strlist_itr_getnext(pitr)) {
      ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--keyring", strlist_itr_getstr(pitr)));
    }
    ERRCHECK(dwErr, strlist_add_node(pOption, "--sign"));
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, "--yes --always-trust --compress-level 0 --symmetric"));
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  if (pOption) strlist_release(pOption);
  if (dwErr != ERROR_SUCCESS) {
    FREE(pCommandLine);
  }
  return pCommandLine;
}

/*------------------------------------------------------------*/
DWORD
gpgez_encrypt_sym(LPCSTR pszGpgHome, PSTRLIST pSecRing, PSTRLIST pPubRing, BOOL fNoSign, HANDLE hStdin, HANDLE hStdout, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess) {
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  LPSTR pCommandLine = NULL;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  ERRCHECK(dwErr, setup_si(&si, &hStdin, &hStdout, &hNull));
  if (!(pCommandLine = gpgez_encrypt_sym_generate_commandline(pszGpgHome, pSecRing, pPubRing, fNoSign, (INT)hCommand, (INT)hStatus))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  pi.hProcess = INVALID_HANDLE_VALUE;
  ERRCHECK(dwErr, rungpg(pCommandLine, &si, &pi));
  CloseHandleClean(&hNull);

  *phProcess = pi.hProcess;

finish_func:
  FREE(pCommandLine);

  return (dwErr);
}

/*------------------------------------------------------------*/
/* For Sign */
/*------------------------------------------------------------*/
static LPSTR
gpgez_sign_generate_commandline(LPCSTR pszGpgHome, PSTRLIST pSecRing, PSTRLIST pPubRing, INT iCommandFd, INT iStatusFd) {
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  PSTRLIST_ITERATOR pitr;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, GPGCOMMANDLINE));
  ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--homedir", pszGpgHome));
  for (pitr = strlist_itr_gethead(pSecRing); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--secret-keyring", strlist_itr_getstr(pitr)));
  }
  for (pitr = strlist_itr_gethead(pPubRing); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--keyring", strlist_itr_getstr(pitr)));
  }
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--command-fd", iCommandFd));
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--status-fd", iStatusFd));
  ERRCHECK(dwErr, strlist_add_node(pOption, "--batch --yes --always-trust --detach-sign"));
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  if (pOption) strlist_release(pOption);
  if (dwErr != ERROR_SUCCESS) {
    FREE(pCommandLine);
  }
  return pCommandLine;
}

/*------------------------------------------------------------*/
DWORD
gpgez_sign(LPCSTR pszGpgHome, PSTRLIST pSecRing, PSTRLIST pPubRing, HANDLE hStdin, HANDLE hStdout, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess) {
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  LPSTR pCommandLine = NULL;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  DWORD dwErr;

  ERRCHECK(dwErr, setup_si(&si, &hStdin, &hStdout, &hNull));
  if (!(pCommandLine = gpgez_sign_generate_commandline(pszGpgHome, pSecRing, pPubRing, (INT)hCommand, (INT)hStatus))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  pi.hProcess = INVALID_HANDLE_VALUE;
  ERRCHECK(dwErr, rungpg(pCommandLine, &si, &pi));
  CloseHandleClean(&hNull);

  *phProcess = pi.hProcess;

finish_func:
  CloseHandleClean(&(pi.hProcess));
  FREE(pCommandLine);

  return (dwErr);
}

/*------------------------------------------------------------*/
/* For Verify */
/*------------------------------------------------------------*/
static LPSTR
gpgez_verify_generate_commandline(LPCSTR pszGpgHome, PSTRLIST pSecRing, PSTRLIST pPubRing, LPCSTR pszSigFile, INT iCommandFd, INT iStatusFd) {
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  PSTRLIST_ITERATOR pitr;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  ERRCHECK(dwErr, strlist_add_node(pOption, GPGCOMMANDLINE));
  ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--homedir", pszGpgHome));
  for (pitr = strlist_itr_gethead(pSecRing); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--secret-keyring", strlist_itr_getstr(pitr)));
  }
  for (pitr = strlist_itr_gethead(pPubRing); pitr; pitr = strlist_itr_getnext(pitr)) {
    ERRCHECK(dwErr, strlist_add_node_ksdq(pOption, "--keyring", strlist_itr_getstr(pitr)));
  }
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--command-fd", iCommandFd));
  ERRCHECK(dwErr, strlist_add_node_kd(pOption, "--status-fd", iStatusFd));
  ERRCHECK(dwErr, strlist_add_node(pOption, "--batch --yes --always-trust --verify"));
  ERRCHECK(dwErr, strlist_add_node_dq(pOption, pszSigFile));
  ERRCHECK(dwErr, strlist_add_node(pOption, "-"));
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  if (pOption) strlist_release(pOption);
  if (dwErr != ERROR_SUCCESS) {
    FREE(pCommandLine);
  }
  return pCommandLine;
}

/*------------------------------------------------------------*/
DWORD
gpgez_verify(LPCSTR pszGpgHome, PSTRLIST pSecRing, PSTRLIST pPubRing, LPCSTR pszSigFile, HANDLE hStdin, HANDLE hCommand, HANDLE hStatus, PHANDLE phProcess) {
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  LPSTR pCommandLine = NULL;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  DWORD dwErr;

  ERRCHECK(dwErr, setup_si(&si, &hStdin, &hNull, &hNull));
  if (!(pCommandLine = gpgez_verify_generate_commandline(pszGpgHome, pSecRing, pPubRing, pszSigFile, (INT)hCommand, (INT)hStatus))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  pi.hProcess = INVALID_HANDLE_VALUE;
  ERRCHECK(dwErr, rungpg(pCommandLine, &si, &pi));
  CloseHandleClean(&hNull);

  *phProcess = pi.hProcess;

finish_func:
  FREE(pCommandLine);

  return (dwErr);
}

/*------------------------------------------------------------*/

