/**
 * runtar.c
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <windows.h>
#include "util.h"
#include "gpgez.h"

/*------------------------------------------------------------*/
static DWORD
runtar(LPCSTR basedir, LPCSTR pCommandLineOptions, STARTUPINFO *psi, PROCESS_INFORMATION *ppi) {
  CHAR szTarCommandFile[MAX_PATH + 1];
  CHAR szHomeDirectory[MAX_PATH + 1];
  DWORD dwErr = ERROR_SUCCESS;

  GetModuleFileName(NULL, szHomeDirectory, MAX_PATH);
  *(strryen(szHomeDirectory) + 1) = '\0';
  STRCPY(szTarCommandFile, szHomeDirectory);
  STRCAT(szTarCommandFile, GPGBINDIR);
  STRCAT(szTarCommandFile, "\\");
  STRCAT(szTarCommandFile, TARCOMMAND);
  if (!CreateProcess(szTarCommandFile, (LPTSTR)pCommandLineOptions, NULL, NULL, TRUE, 0, NULL, basedir, psi, ppi)) {
    dwErr = GetLastError();
    goto finish_func;
  }
  CloseHandleClean(&(ppi->hThread));
finish_func:
  return dwErr;
}

/*------------------------------------------------------------*/
DWORD
gpgez_tar(LPCSTR basedir, PSTRLIST files, HANDLE hStdOut, HANDLE hStdErr, PHANDLE phProcess) {
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  PSTRLIST_ITERATOR pitr;
  HANDLE hNull = INVALID_HANDLE_VALUE;
  SECURITY_ATTRIBUTES sa;
  DWORD dwErr = ERROR_SUCCESS;

  ZeroMemory(&sa, sizeof(SECURITY_ATTRIBUTES));
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.bInheritHandle = TRUE;
  hNull = CreateFile ("nul",
                      GENERIC_READ|GENERIC_WRITE,
                      FILE_SHARE_READ|FILE_SHARE_WRITE,
                      &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

  ZeroMemory(&si, sizeof(STARTUPINFO));
  si.cb = sizeof(STARTUPINFO);
  si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
  si.wShowWindow = SW_HIDE;
  si.hStdInput = hNull;
  si.hStdOutput = hStdOut;
  si.hStdError = hStdErr;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  if ((dwErr = strlist_add_node(pOption, " -C")) != ERROR_SUCCESS) goto finish_func;
  if ((dwErr = strlist_add_node_dq(pOption, basedir)) != ERROR_SUCCESS) goto finish_func;
  if ((dwErr = strlist_add_node(pOption, "-czvf - ")) != ERROR_SUCCESS) goto finish_func;
  for (pitr = strlist_itr_gethead(files); pitr; pitr = strlist_itr_getnext(pitr)) {
    if ((dwErr = strlist_add_node_dq(pOption, strlist_itr_getstr(pitr))) != ERROR_SUCCESS) goto finish_func;
  }
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  pi.hProcess = INVALID_HANDLE_VALUE;
  if ((dwErr = runtar(basedir, pCommandLine, &si, &pi)) != ERROR_SUCCESS) goto finish_func;

  *phProcess = pi.hProcess;

finish_func:
  if (hNull != INVALID_HANDLE_VALUE) CloseHandleClean(&hNull);
  if (pOption) strlist_release(pOption);
  FREE(pCommandLine);
  return dwErr;
}

/*------------------------------------------------------------*/
DWORD
gpgez_untar(LPCSTR basedir, HANDLE hStdIn, HANDLE hStdErr, BOOL fOverwrite, PHANDLE phProcess) {
  SECURITY_ATTRIBUTES sa;
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  LPSTR pCommandLine = NULL;
  PSTRLIST pOption = NULL;
  DWORD dwErr = ERROR_SUCCESS;
  HANDLE hNull = INVALID_HANDLE_VALUE;

  ZeroMemory(&sa, sizeof(SECURITY_ATTRIBUTES));
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.bInheritHandle = TRUE;
  hNull = CreateFile ("nul",
                      GENERIC_READ|GENERIC_WRITE,
                      FILE_SHARE_READ|FILE_SHARE_WRITE,
                      &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

  ZeroMemory(&si, sizeof(STARTUPINFO));
  si.cb = sizeof(STARTUPINFO);
  si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
  si.wShowWindow = SW_HIDE;
  si.hStdInput = hStdIn;
  si.hStdOutput = hNull;
  si.hStdError = hStdErr;

  if (!(pOption = strlist_create())) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  if ((dwErr = strlist_add_node(pOption, " -C")) != ERROR_SUCCESS) goto finish_func;
  if ((dwErr = strlist_add_node_dq(pOption, basedir)) != ERROR_SUCCESS) goto finish_func;
  if (fOverwrite) {
    if ((dwErr = strlist_add_node(pOption, "-xvf ")) != ERROR_SUCCESS) goto finish_func;
  } else {
    if ((dwErr = strlist_add_node(pOption, "-xvkf ")) != ERROR_SUCCESS) goto finish_func;
  }
  if ((dwErr = strlist_add_node(pOption, "-")) != ERROR_SUCCESS) goto finish_func;
  if (!(pCommandLine = strlist_concatenate(pOption, " "))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  pi.hProcess = INVALID_HANDLE_VALUE;
  if ((dwErr = runtar(basedir, pCommandLine, &si, &pi)) != ERROR_SUCCESS) goto finish_func;

  *phProcess = pi.hProcess;

finish_func:
  if (hNull != INVALID_HANDLE_VALUE) CloseHandleClean(&hNull);
  if (pOption) strlist_release(pOption);
  FREE(pCommandLine);
  return dwErr;
}

/*------------------------------------------------------------*/
