/**
 * selectKey.c
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include "util.h"
#include "gpgez.h"
#include "JanusHS.h"
#include "resource.h"

extern HINSTANCE g_hInstance;

typedef struct janus_selectkey {
  LPCSTR szGpgHome;
  PSTRLIST pslSecRing;
  PSTRLIST pslPubRing;
  PSTRLIST pslSelected;
  LPDWORD lpdwEncType;
} JANUS_SELECTKEY, *PJANUS_SELECTKEY;

/*------------------------------------------------------------*/
static DWORD
selectkey_show(HWND hwnd) {
  HANDLE hReadP = INVALID_HANDLE_VALUE;
  HANDLE hStdoutC = INVALID_HANDLE_VALUE;
  HANDLE hProcess = INVALID_HANDLE_VALUE;
  CHAR szLine[MAX_LINE + 1];
  CHAR szUid[MAX_LINE + 1];
  CHAR szFpr[MAX_LINE + 1];
  UINT i;
  LPSTR ps, pe;
  LV_ITEM item;
  LPSTR psjis = NULL;
  UINT iLine = 0;
  DWORD dwErr = ERROR_SUCCESS;
  HWND hwndList;
  PJANUS_SELECTKEY pjsk;

  pjsk = (PJANUS_SELECTKEY)GetWindowLong(hwnd, GWL_USERDATA);
  hwndList = GetDlgItem(hwnd, IDC_LIST_KEYS);
  szUid[0] = '\0';
  szFpr[0] = '\0';
  if (!ListView_DeleteAllItems(hwndList)) {
    dwErr = GetLastError();
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }

  dwErr = CreateReadPipe(&hReadP, &hStdoutC);
  if (dwErr != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  dwErr = gpgez_listkeys(pjsk->szGpgHome, pjsk->pslPubRing, FALSE, hStdoutC, &hProcess);
  if (dwErr != ERROR_SUCCESS) {
    DEBUGERROR(0, dwErr);
    goto finish_func;
  }
  CloseHandleClean(&hStdoutC);

  while (dwErr == ERROR_SUCCESS) {
    dwErr = ReadPipeLine(&hReadP, szLine, MAX_LINE);
    if ((dwErr != ERROR_SUCCESS) && (dwErr != ERROR_HANDLE_EOF)) {
      DEBUGERROR(0, dwErr);
      goto finish_func;
    }
    if (!STRNCMPL("pub:", szLine)) {
      if (szUid[0] || szFpr[0]) {
        /* ListView ɒǉ */
        item.mask = LVIF_TEXT;
        item.pszText = szUid;
        item.iItem = iLine;
        item.iSubItem = 0;
        ListView_InsertItem(hwndList, &item);

        item.pszText = szFpr;
        item.iItem = iLine;
        item.iSubItem = 1;
        ListView_SetItem(hwndList, &item);

        szUid[0] = '\0';
        szFpr[0] = '\0';
        iLine++;
      }
      ps = szLine;
      for (i = 0; i < 9; i++) {
        if (!(ps = STRCHR(ps, ':'))) {
          dwErr = ERROR_INVALID_DATA;
          DEBUGERROR(0, dwErr);
          goto finish_func;
        }
        ps++;
      }
      if (!(pe = STRCHR(ps, ':'))) {
        dwErr = ERROR_INVALID_DATA;
        DEBUGERROR(0, dwErr);
        goto finish_func;
      }
      STRNCPY(szUid, ps, pe - ps);
      szUid[pe - ps] = '\0';
      if (!(psjis = utf82sjis(szUid))) {
        dwErr = ERROR_OUTOFMEMORY;
        DEBUGERROR(0, dwErr);
        goto finish_func;
      }
      STRCPY(szUid, psjis);
      FREE(psjis);
    }
    if ((!STRNCMPL("fpr:", szLine)) && !szFpr[0]) {
      ps = szLine;
      for (i = 0; i < 9; i++) {
        if (!(ps = STRCHR(ps, ':'))) {
          dwErr = ERROR_INVALID_DATA;
          DEBUGERROR(0, dwErr);
          goto finish_func;
        }
        ps++;
      }
      if (!(pe = STRCHR(ps, ':'))) {
        dwErr = ERROR_INVALID_DATA;
        DEBUGERROR(0, dwErr);
        goto finish_func;
      }
      STRNCPY(szFpr, ps, pe - ps);
      szFpr[pe - ps] = '\0';
    }
    if (dwErr == ERROR_HANDLE_EOF) {
      if (szFpr[0]) {
        item.mask = LVIF_TEXT;
        item.pszText = szUid;
        item.iItem = iLine;
        item.iSubItem = 0;
        ListView_InsertItem(hwndList, &item);

        item.pszText = szFpr;
        item.iItem = iLine;
        item.iSubItem = 1;
        ListView_SetItem(hwndList, &item);

        szUid[0] = '\0';
        szFpr[0] = '\0';
        iLine++;
      }
      dwErr = ERROR_SUCCESS;
      break;
    }
  }
finish_func:
  CloseHandleClean(&hProcess);
  CloseHandleClean(&hReadP);
  CloseHandleClean(&hStdoutC);
  return dwErr;
}

/*------------------------------------------------------------*/
static BOOL
selectkey_Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam) {
  LV_COLUMN lvcol;

  SetWindowLong(hwnd, GWL_USERDATA, (LONG)lParam);
  EnableWindow(GetDlgItem(hwnd, IDC_LIST_KEYS), TRUE);
  EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_KEYMANAGER), TRUE);
  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_PASS1), FALSE);
  EnableWindow(GetDlgItem(hwnd, IDC_EDIT_PASS2), FALSE);
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_PASS1), MAX_LINE);
  Edit_LimitText(GetDlgItem(hwnd, IDC_EDIT_PASS2), MAX_LINE);
  Button_SetCheck(GetDlgItem(hwnd, IDC_RADIO_PUB), BST_CHECKED);

  lvcol.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
  lvcol.fmt = LVCFMT_LEFT;
  lvcol.cx = 200;
  lvcol.pszText = "Name";
  lvcol.iSubItem = 0;
  ListView_InsertColumn(GetDlgItem(hwnd, IDC_LIST_KEYS), 0, &lvcol);

  lvcol.cx = 300;
  lvcol.pszText = "Fingerprint";
  lvcol.iSubItem = 1;
  ListView_InsertColumn(GetDlgItem(hwnd, IDC_LIST_KEYS), 1, &lvcol);
  return TRUE;
}

/*------------------------------------------------------------*/
static void
selectkey_Cls_OnShowWindow(HWND hwnd, BOOL fShow, UINT status) {
  selectkey_show(hwnd);
}

/*------------------------------------------------------------*/
static VOID
selectkey_Cls_OnClose(HWND hwnd) {
  PJANUS_SELECTKEY pjsk;

  pjsk = (PJANUS_SELECTKEY)GetWindowLong(hwnd, GWL_USERDATA);
  *(pjsk->lpdwEncType) = ENCTYPE_NONE;
  EndDialog(hwnd, (INT)ERROR_CANCELLED);
}

/*------------------------------------------------------------*/
static VOID
selectkey_Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify) {
  switch(id) {
  case IDOK:
    {
      PJANUS_SELECTKEY pjsk;

      pjsk = (PJANUS_SELECTKEY)GetWindowLong(hwnd, GWL_USERDATA);
      if (Button_GetCheck(GetDlgItem(hwnd, IDC_RADIO_PUB))) {
        INT iNext = -1;
        HWND hwndKeyList;
        LVITEM item;
        CHAR szLine[MAX_LINE + 1];
        DWORD dwErr = ERROR_SUCCESS;

        item.mask = LVIF_TEXT;
        item.pszText = szLine;
        item.cchTextMax = MAX_LINE;
        hwndKeyList = GetDlgItem(hwnd, IDC_LIST_KEYS);
        while (1) {
          iNext = ListView_GetNextItem(hwndKeyList, iNext, LVNI_ALL | LVNI_SELECTED);
          if (iNext == -1) break;
          item.iItem = iNext;
          item.iSubItem = 1;
          ListView_GetItem(hwndKeyList, &item);
          if ((dwErr = strlist_add_node(pjsk->pslSelected, item.pszText)) != ERROR_SUCCESS) {
            DEBUGERROR(0, dwErr);
            EndDialog(hwnd, (INT)dwErr);
          }
        }
        if (bufflist_getcount(pjsk->pslSelected) == 0) {
          CHAR m_buff[MAX_PATH + 1];

          if (LoadString(g_hInstance, IDS_ERROR_NORECIPIENTSSELECTED, m_buff, MAX_PATH) == 0) {
            STRCPY(m_buff, "No Recipients Selected.\nAborting");
          }
          if (IDNO == MessageBox(hwnd, m_buff, "Confirmation", MB_YESNO | MB_ICONQUESTION)) {
            break;
          }
        }
        *(pjsk->lpdwEncType) = ENCTYPE_PUBLIC;
        EndDialog(hwnd, (INT)ERROR_SUCCESS);
      } else {
        CHAR szLine1[MAX_LINE + 1];
        CHAR szLine2[MAX_LINE + 1];
        DWORD dwErr = ERROR_SUCCESS;

        Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_PASS1), szLine1, MAX_LINE);
        Edit_GetText(GetDlgItem(hwnd, IDC_EDIT_PASS2), szLine2, MAX_LINE);
        if (STRCMP(szLine1, szLine2)) {
          CHAR m_buff[MAX_PATH + 1];

          if (LoadString(g_hInstance, IDS_ERROR_PASSPHRASENOMATCH, m_buff, MAX_PATH) == 0) {
            STRCPY(m_buff, "Passphrases Miss Match!");
          }
          MessageBox(hwnd, m_buff, "Error", MB_OK);
          break;
        }
        if (STRLEN(szLine1) < 4) {
          CHAR m_buff[MAX_PATH + 1];

          if (LoadString(g_hInstance, IDS_ERROR_PASSPHRASETOOSHORT, m_buff, MAX_PATH) == 0) {
            STRCPY(m_buff, "Passphrase too short");
          }
          MessageBox(hwnd, m_buff, "Error", MB_OK);
          break;
        }
        if ((dwErr = strlist_add_node(pjsk->pslSelected, szLine1)) != ERROR_SUCCESS) {
          DEBUGERROR(0, dwErr);
          EndDialog(hwnd, (INT)dwErr);
        }
        *(pjsk->lpdwEncType) = ENCTYPE_SYMMETRIC;
        EndDialog(hwnd, (INT)ERROR_SUCCESS);
      }
    }
    break;
  case IDCANCEL:
    {
      PJANUS_SELECTKEY pjsk;

      pjsk = (PJANUS_SELECTKEY)GetWindowLong(hwnd, GWL_USERDATA);
      *(pjsk->lpdwEncType) = ENCTYPE_NONE;
      EndDialog(hwnd, (INT)ERROR_CANCELLED);
    }
    break;
  case IDC_RADIO_PUB:
    EnableWindow(GetDlgItem(hwnd, IDC_LIST_KEYS), TRUE);
    EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_KEYMANAGER), TRUE);
    EnableWindow(GetDlgItem(hwnd, IDC_EDIT_PASS1), FALSE);
    EnableWindow(GetDlgItem(hwnd, IDC_EDIT_PASS2), FALSE);
    break;
  case IDC_RADIO_SYM:
    EnableWindow(GetDlgItem(hwnd, IDC_LIST_KEYS), FALSE);
    EnableWindow(GetDlgItem(hwnd, IDC_BUTTON_KEYMANAGER), FALSE);
    EnableWindow(GetDlgItem(hwnd, IDC_EDIT_PASS1), TRUE);
    EnableWindow(GetDlgItem(hwnd, IDC_EDIT_PASS2), TRUE);
    break;
  case IDC_BUTTON_KEYMANAGER:
    {
      DWORD dwErr = ERROR_SUCCESS;
      PJANUS_SELECTKEY pjsk;

      pjsk = (PJANUS_SELECTKEY)GetWindowLong(hwnd, GWL_USERDATA);
      dwErr = editGpgKey(hwnd, pjsk->szGpgHome, pjsk->pslSecRing, pjsk->pslPubRing);
      if (dwErr != ERROR_SUCCESS) {
        EndDialog(hwnd, (INT)dwErr);
      }
      selectkey_show(hwnd);
    }
    break;
  }
}

/*------------------------------------------------------------*/
BOOL CALLBACK
procDlgSelectKey(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    HANDLE_MSG(hDlg, WM_INITDIALOG, selectkey_Cls_OnInitDialog);
    HANDLE_MSG(hDlg, WM_SHOWWINDOW, selectkey_Cls_OnShowWindow);
    HANDLE_MSG(hDlg, WM_CLOSE, selectkey_Cls_OnClose);
    HANDLE_MSG(hDlg, WM_COMMAND, selectkey_Cls_OnCommand);
  case WM_NOTIFY:
    if (wParam == IDC_LIST_KEYS) {
      LPNMHDR lpnmh = (LPNMHDR)lParam;
      if (lpnmh->code == NM_DBLCLK) {
        SendMessage(hDlg, WM_COMMAND, (WPARAM)IDOK, (LPARAM)NULL);
        return TRUE;
      }
    }
    return FALSE;
  default:
    return FALSE;
  }
  return TRUE;
}

/*------------------------------------------------------------*/
DWORD
selectGpgKey(HWND hwnd, LPCSTR szGpgHome, PSTRLIST pslSecRing, PSTRLIST pslPubRing, PSTRLIST pslSelected, LPDWORD lpdwEncType) {
  JANUS_SELECTKEY jsk;
  DWORD dwErr = ERROR_SUCCESS;

  jsk.szGpgHome = szGpgHome;
  jsk.pslSecRing = pslSecRing;
  jsk.pslPubRing = pslPubRing;
  jsk.pslSelected = pslSelected;
  jsk.lpdwEncType = lpdwEncType;
  dwErr = (DWORD) DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_DIALOG_SELKEY), hwnd, procDlgSelectKey, (LPARAM)&jsk);
  return dwErr;
}

/*------------------------------------------------------------*/

