/**
 * util.c
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "util.h"
#include <windowsx.h>

/*------------------------------------------------------------------------*/
VOID
strtrim(LPSTR str) {
  LPSTR src;
  LPSTR dst;

  for (src = str; *src == ' '; src++)
    ;
  for (dst = str; *dst; dst++)
    ;
  dst--;
  for (; dst > str && *dst == ' '; dst--) {
    *dst = '\0';
  }
  return;
}

/*------------------------------------------------------------------------*/
LPSTR
strgetat(LPCSTR pLine, WORD iSep, INT iCnt) {
  PCSTR ps, pe;
  PSTR pStr;
  INT i;

  ps = pLine;
  for (i = 0; i < iCnt; i++) {
    ps = STRCHR(ps, iSep);
    if (!ps) {
      pStr = NULL;
      goto finish_func;
    }
    ps++;
  }
  pe = STRCHR(ps, iSep);
  if (!pe) {
    pe = STRCHR(ps, '\0');
  }
  pStr = (LPSTR)MALLOC(pe - ps + 1);
  if (!pStr) {
    goto finish_func;
  }
  STRNCPY(pStr, ps, pe - ps);
  *(pStr + (pe - ps)) = '\0';
finish_func:
  return pStr;
}

/*------------------------------------------------------------------------*/
LPSTR
escapechar(LPSTR pstr, CHAR toEscaped, CHAR escape) {
  DWORD cnt = 0, len = 0;
  LPSTR p, p1, pnew;

  for (p = pstr; *p; p++) {
    if (*p == toEscaped) cnt++;
  }
  if ((pnew = (LPSTR)MALLOC(STRLEN(pstr) + cnt + 1)) == NULL) {
    goto finish_func;
  }
  for (p = pstr, p1 = pnew; *p; p++) {
    if (*p == toEscaped) {
      *p1 = escape;
      p1++;
    }
    *p1 = *p;
    p1++;
  }
  *p1 = '\0';
finish_func:
  return pnew;
}

/*------------------------------------------------------------------------*/
LPSTR
_ustrchr(LPCSTR str, WORD w) {
  LPCSTR p;
  LPSTR pos = NULL;

  for (p = str; *p; p++) {
    if (*p == w) {
      pos = (LPSTR)p;
      break;
    }
  }
  return pos;
}

/*------------------------------------------------------------------------*/
LPSTR
_ustrdup(LPCSTR str) {
  LPSTR s;

  s = (LPSTR)MALLOC(STRLEN(str) + 1);
  if (s) {
    STRCPY(s, str);
  }
  return s;
}
/*------------------------------------------------------------------------*/
LPSTR
_ustrncpy(LPSTR s1, LPCSTR s2, INT n) {
  INT iCnt = 0;
  LPSTR ps1 = s1;
  LPCSTR ps2 = s2;

  while (iCnt <= n) {
    *ps1 = *ps2;
    if (!*ps2) break;
    ps1++;
    ps2++;
    iCnt++;
  }
  return s1;
}

/*------------------------------------------------------------------------*/
INT
_ustrncmp(LPCSTR s1, LPCSTR s2, INT n) {
  INT iCnt = 0;
  LPCSTR ps1 = s1, ps2 = s2;

  while (1) {
    if (iCnt >= n) break;
    if (*ps1 > *ps2) return 1;
    if (*ps1 < *ps2) return -1;
    if ((*ps1 == '\0') && (*ps2 == '\0')) break;
    ps1++; ps2++;
    iCnt++;
  }
  return 0;
}

/*------------------------------------------------------------------------*/
LPSTR
strryen(LPCSTR str) {
  LPCSTR p;
  LPSTR pos = NULL;

  for (p = str; *p; p++) {
    if (*p == '\\') pos = (LPSTR) p;
    if (IsDBCSLeadByte(*p)) {
      p++;
    }
  }
  return pos;
}

/*------------------------------------------------------------------------*/
DWORD
strsplitgetat(LPSTR pstr, INT i, DWORD pos, LPSTR pbuff, INT max) {
  DWORD cnt = 0;
  LPSTR ps;
  LPSTR pe;
  INT len;

  *pbuff = '\0';
  ps = pstr;
  while (cnt < 1) {
    while (*ps && (*ps != i))
      ps++;
    if (!*ps) {
      return ERROR_INVALID_DATA;
    }
    cnt++;
  }
  ps++;
  for (pe = ps; *pe && (*pe != i); pe++)
    ;
  len = ps - pe > max ? max : ps - pe;
  STRNCPY(pbuff, ps, len);
  if (len < max) {
    *(pbuff + len) = '\0';
  }
  return ERROR_SUCCESS;
}

/*------------------------------------------------------------------------*/
LPSTR
sjis2utf8(LPSTR psrc) {
  int i;
  LPWSTR pwcs = NULL;
  LPSTR putf8 = NULL;

  i = MultiByteToWideChar(CP_ACP, 0, psrc, -1, NULL, 0);
  if (i == 0) goto finish_func;
  pwcs = (LPWSTR)MALLOC(sizeof(WCHAR) * (i + 1));
  if (!pwcs) goto finish_func;
  i = MultiByteToWideChar(CP_ACP, 0, psrc, -1, pwcs, i);
  if (i == 0) goto finish_func;
  *(pwcs + i) = 0;

  i = WideCharToMultiByte(CP_UTF8, 0, pwcs, -1, NULL, 0, NULL, NULL);
  if (i == 0) goto finish_func;
  putf8 = (LPSTR)MALLOC(sizeof(CHAR) * (i + 1));
  if (!putf8) goto finish_func;
  i = WideCharToMultiByte(CP_UTF8, 0, pwcs, -1, putf8, i, NULL, NULL);
  if (i == 0) goto finish_func;
  FREE(pwcs);
  pwcs = NULL;
finish_func:
  if (pwcs) FREE(pwcs);
  return putf8;
}

/*------------------------------------------------------------------------*/
LPSTR
utf82sjis(LPSTR psrc) {
  int i;
  LPWSTR pwcs = NULL;
  LPSTR psjis = NULL;

  i = MultiByteToWideChar(CP_UTF8, 0, psrc, -1, NULL, 0);
  if (i == 0) goto finish_func;
  pwcs = (LPWSTR)MALLOC(sizeof(WCHAR) *(i + 1));
  if (!pwcs) goto finish_func;
  i = MultiByteToWideChar(CP_UTF8, 0, psrc, -1, pwcs, i);
  if (i == 0) goto finish_func;
  *(pwcs + i) = 0;

  i = WideCharToMultiByte(CP_ACP, 0, pwcs, -1, NULL, 0, NULL, NULL);
  if (i == 0) goto finish_func;
  psjis = (LPSTR)MALLOC(sizeof(CHAR) * (i + 1));
  if (!psjis) goto finish_func;
  i = WideCharToMultiByte(CP_ACP, 0, pwcs, -1, psjis, i, NULL, NULL);
  if (i == 0) goto finish_func;
  FREE(pwcs);
  pwcs = NULL;
finish_func:
  if (pwcs) FREE(pwcs);
  return psjis;
}

/*------------------------------------------------------------------------*/
LPSTR
unescapeutf82sjis(LPCSTR src) {
  LPSTR putf8 = NULL, psjis = NULL, p;
  LPCSTR p1;

  putf8 = (LPSTR)MALLOC(STRLEN(src) + 1);
  if (!putf8) return NULL;
  for (p1 = src, p = putf8; *p1; p1++, p++) {
    if (*p1 == '%') {
      *p = (*(p1 + 1) - '0') << 4 | (*(p1 + 2) - '0');
      p += 2;
    } else {
      *p = *p1;
    }
  }
  *p = '\0';
  psjis = utf82sjis(putf8);
  FREE(putf8);
  return psjis;
}

/*------------------------------------------------------------------------*/
LPSTR
emailgetname(LPCSTR pemail) {
  LPCSTR ps = pemail, pe = NULL;
  LPSTR pstr = NULL;
  DWORD dwErr = ERROR_SUCCESS;

  for (pe = ps; *pe; pe++) {
    if ((*pe == '(') || (*pe == '<')) break;
  }
  if (!*pe) {
    if ((pstr = STRDUP("")) == NULL) {
      dwErr = ERROR_OUTOFMEMORY;
      goto finish_func;
    }
  } else {
    if ((pstr = (LPSTR)MALLOC(sizeof(CHAR) * (pe - ps + 1))) == NULL) {
      dwErr = ERROR_OUTOFMEMORY;
      goto finish_func;
    }
    STRNCPY(pstr, ps, pe - ps);
    *(pstr + (pe - ps)) = '\0';
  }
finish_func:
  return pstr;
}

/*------------------------------------------------------------------------*/
LPSTR
emailgetemail(LPCSTR pemail) {
  LPCSTR ps = pemail, pe = NULL;
  LPSTR pstr = NULL;
  DWORD dwErr = ERROR_SUCCESS;

  for (ps = pemail; *ps; ps++) {
    if (*ps == '<') {
      ps++;
      for (pe = ps; *pe; pe++) {
        if (*pe == '>') {
          if ((pstr = (LPSTR)MALLOC(sizeof(CHAR) * (pe - ps + 1))) == NULL) {
            dwErr = ERROR_OUTOFMEMORY;
            goto finish_func;
          }
          STRNCPY(pstr, ps, pe - ps);
          *(pstr + (pe - ps)) = '\0';
          dwErr = ERROR_SUCCESS;
          goto finish_func;
        }
      }
      dwErr = ERROR_INVALID_DATA;
      goto finish_func;
    }
  }
  ps = pemail;
  for (pe = pemail; *pe; pe++) {
    if (*pe == '(') break;
  }
  if ((pstr = (LPSTR)MALLOC(sizeof(CHAR) * (pe - ps + 1))) == NULL) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  STRNCPY(pstr, ps, pe - ps);
  *(pstr + (pe - ps)) = '\0';
finish_func:
  return pstr;
}

/*------------------------------------------------------------------------*/
LPSTR
emailgetcomment(LPCSTR pemail) {
  LPCSTR ps = pemail, pe = NULL;
  LPSTR pstr = NULL;
  DWORD dwErr = ERROR_SUCCESS;

  for (ps = pemail; *ps; ps++) {
    if (*ps == '(') {
      ps++;
      for (pe = ps; *pe; pe++) {
        if (*pe == ')') {
          if ((pstr = (LPSTR)MALLOC(sizeof(CHAR) * (pe - ps + 1))) == NULL) {
            dwErr = ERROR_OUTOFMEMORY;
            goto finish_func;
          }
          STRNCPY(pstr, ps, pe - ps);
          *(pstr + (pe - ps)) = '\0';
          dwErr = ERROR_SUCCESS;
          goto finish_func;
        }
      }
      dwErr = ERROR_INVALID_DATA;
      goto finish_func;
    }
  }
  if ((pstr = STRDUP("")) == NULL) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
finish_func:
  return pstr;
}

/*------------------------------------------------------------------------*/
/* Xgp֐ */
/*------------------------------------------------------------------------*/
PBUFFLIST
bufflist_create(LPVOID (*clone_buff)(PVOID src), VOID (*delete_buff)(PVOID buff)) {
  PBUFFLIST plist;

  if (!(plist = (PBUFFLIST)MALLOC(sizeof(BUFFLIST)))) {
    goto finish_func;
  }
  plist->clone_buff = clone_buff;
  plist->delete_buff = delete_buff;
  plist->phead = NULL;
finish_func:
  return plist;
}

/*------------------------------------------------------------------------*/
VOID
bufflist_release(PBUFFLIST plist) {
  PBUFFLIST_NODE pnode, pnode_tmp;

  pnode = plist->phead;
  while (pnode) {
    pnode_tmp = pnode->next;
    plist->delete_buff(pnode->buff);
    FREE(pnode);
    pnode = pnode_tmp;
  }
  FREE(plist);
  return;
}

/*------------------------------------------------------------------------*/
DWORD
bufflist_add_node_head(PBUFFLIST plist, LPVOID buff) {
  PBUFFLIST_NODE pnode = NULL;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pnode = (PBUFFLIST_NODE)MALLOC(sizeof(BUFFLIST_NODE)))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  if (!(pnode->buff = plist->clone_buff(buff))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  pnode->next = plist->phead;
  plist->phead = pnode;
finish_func:
  if (dwErr != ERROR_SUCCESS) {
    if (pnode->buff) (plist->delete_buff)(pnode->buff);
    free(pnode);
  }
  return dwErr;
}

/*------------------------------------------------------------------------*/
DWORD
bufflist_add_node_tail(PBUFFLIST plist, LPVOID buff) {
  PBUFFLIST_NODE pnode = NULL, ptmp;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pnode = (PBUFFLIST_NODE)MALLOC(sizeof(BUFFLIST_NODE)))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  if (!(pnode->buff = plist->clone_buff(buff))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  pnode->next = NULL;
  if (plist->phead) {
    for (ptmp = plist->phead; ptmp->next; ptmp = ptmp->next)
      ;
    ptmp->next = pnode;
  } else {
    plist->phead = pnode;
  }
finish_func:
  if (dwErr != ERROR_SUCCESS) {
    if (pnode->buff) (plist->delete_buff)(pnode->buff);
    FREE(pnode);
  }
  return dwErr;
}

/*------------------------------------------------------------------------*/
DWORD
bufflist_add_node_sorted(PBUFFLIST plist, LPVOID buff, INT (*compare)(LPVOID pbuff1, LPVOID pbuff2)) {
  PBUFFLIST_NODE pnode = NULL, ptmp;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pnode = (PBUFFLIST_NODE)MALLOC(sizeof(BUFFLIST_NODE)))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  if (!(pnode->buff = plist->clone_buff(buff))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  pnode->next = NULL;
  if (plist->phead) {
    for (ptmp = plist->phead; compare(buff, ptmp->next->buff) <= 0; ptmp = ptmp->next) {
      if (ptmp->next == NULL) break;
    }
    pnode->next = ptmp->next;
    ptmp->next = pnode;
  } else {
    plist->phead = pnode;
  }
finish_func:
  if (dwErr != ERROR_SUCCESS) {
    if (pnode->buff) (plist->delete_buff)(pnode->buff);
    FREE(pnode);
  }
  return dwErr;
}

/*------------------------------------------------------------------------*/
PBUFFLIST
bufflist_clone(PBUFFLIST psrc) {
  PBUFFLIST pdst;
  PBUFFLIST_NODE pnode;
  DWORD dwErr = ERROR_SUCCESS;

  if (!(pdst = bufflist_create(psrc->clone_buff, psrc->delete_buff))) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }
  for (pnode = psrc->phead; pnode; pnode->next) {
    if ((dwErr = (bufflist_add_node_tail(pdst, pnode->buff))) != ERROR_SUCCESS) {
      goto finish_func;
    }
  }
finish_func:
  if (dwErr != ERROR_SUCCESS) {
    if (pdst) bufflist_release(pdst);
    pdst = NULL;
  }
  return pdst;
}

/*------------------------------------------------------------------------*/
INT
bufflist_getcount(PBUFFLIST plist) {
  INT i = 0;
  PBUFFLIST_NODE pnode;

  for (pnode = plist->phead; pnode; pnode = pnode->next)
    i++;
  return i;
}

/*------------------------------------------------------------------------*/
/* ̃Xgp֐ */
/*------------------------------------------------------------------------*/
PVOID
strlist_clone_buff(LPVOID src) {
  LPSTR dst = NULL;

  dst = STRDUP((LPCSTR)src);
  return (LPVOID)dst;
}

/*------------------------------------------------------------------------*/
VOID
strlist_delete_buff(LPVOID buff) {
  FREE(buff);
  return;
}

/*------------------------------------------------------------------------*/
DWORD
strlist_add_node(PSTRLIST plist, LPCSTR str) {
  return bufflist_add_node_tail(plist, (LPVOID)str);
}

/*------------------------------------------------------------------------*/
PSTRLIST
strlist_create(VOID) {
  return (PSTRLIST) bufflist_create(strlist_clone_buff, strlist_delete_buff);
}

/*------------------------------------------------------------------------*/
VOID
strlist_release(PSTRLIST plist) {
  bufflist_release((PBUFFLIST)plist);
}

/*------------------------------------------------------------------------*/
PSTR
strlist_get_node(PSTRLIST plist, INT i) {
  PBUFFLIST_NODE pnode = plist->phead;
  INT ii;

  for (ii = 0; ii < i; ii++) {
    pnode = pnode->next;
    if (!pnode) return NULL;
  }
  return (PSTR)pnode->buff;
}

/*------------------------------------------------------------------------*/
PSTR
strlist_concatenate(PSTRLIST plist, LPCSTR pSep) {
  PBUFFLIST_NODE pnode;
  LPSTR str;
  INT len = 0;

  for (pnode = plist->phead; pnode; pnode = pnode->next) {
    len += lstrlen((LPSTR)(pnode->buff));
    if (pnode->next) len += lstrlen(pSep);
  }
  if (!(str = (LPSTR)MALLOC(len + 1))) {
    goto finish_func;
  }
  *str = '\0';
  for (pnode = plist->phead; pnode; pnode = pnode->next) {
    STRCAT(str, (LPSTR)(pnode->buff));
    if (pnode->next) STRCAT(str, pSep);
  }
finish_func:
  return str;
}

/*------------------------------------------------------------------------*/
PSTRLIST_ITERATOR
strlist_itr_gethead(PSTRLIST plist) {
  return plist->phead;
}

/*------------------------------------------------------------------------*/
PSTRLIST_ITERATOR
strlist_itr_getnext(PSTRLIST_ITERATOR pitr) {
  return pitr->next;
}

/*------------------------------------------------------------------------*/
PSTR
strlist_itr_getstr(PSTRLIST_ITERATOR pitr) {
  return (PSTR) pitr->buff;
}

/*------------------------------------------------------------------------*/
/* t@Cp֐ */
/*------------------------------------------------------------------------*/
DWORD
CreateWritePipe(PHANDLE phPWrite, PHANDLE phCRead) {
  SECURITY_ATTRIBUTES sa;
  HANDLE hPW0 = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;

  if (!CreatePipe(phCRead, &hPW0, &sa, 0)) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if (!DuplicateHandle(GetCurrentProcess(), hPW0, GetCurrentProcess(), phPWrite, 0, FALSE, DUPLICATE_SAME_ACCESS)) {
    dwErr = GetLastError();
    goto finish_func;
  }
finish_func:
  if (hPW0 != INVALID_HANDLE_VALUE) CloseHandleClean(&hPW0);
  return dwErr;
}

/*------------------------------------------------------------------------*/
DWORD
CreateReadPipe(PHANDLE phPRead, PHANDLE phCWrite) {
  SECURITY_ATTRIBUTES sa;
  HANDLE hPR0 = INVALID_HANDLE_VALUE;
  DWORD dwErr = ERROR_SUCCESS;

  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;

  if (!CreatePipe(&hPR0, phCWrite, &sa, 0)) {
    dwErr = GetLastError();
    goto finish_func;
  }
  if (!DuplicateHandle(GetCurrentProcess(), hPR0, GetCurrentProcess(), phPRead, 0, FALSE, DUPLICATE_SAME_ACCESS)) {
    dwErr = GetLastError();
    goto finish_func;
  }
finish_func:
  if (hPR0 != INVALID_HANDLE_VALUE) CloseHandleClean(&hPR0);
  return dwErr;
}

/*------------------------------------------------------------------------*/
DWORD
CloseHandleClean(PHANDLE phFile) {
  DWORD dwErr = ERROR_SUCCESS;

  if (!CloseHandle(*phFile)) {
    dwErr = GetLastError();
  }
  *phFile = INVALID_HANDLE_VALUE;
  return dwErr;
}

/*------------------------------------------------------------------------*/
DWORD
ReadPipeLine(PHANDLE phRead, PSTR pszLine, DWORD dwMax) {
  CHAR c;
  DWORD dwRead;
  DWORD dwLen;
  DWORD dwErr = ERROR_SUCCESS;

  dwLen = 0;
  *pszLine = '\0';
  while (TRUE) {
    if (!ReadFile(*phRead, &c, 1, &dwRead, NULL)) {
      if ((dwErr = GetLastError()) == ERROR_BROKEN_PIPE) {
        *(pszLine + dwLen) = '\0';
        dwErr = ERROR_HANDLE_EOF;
      }
      break;
    }
    if (dwRead > 0) {
      if (dwLen < dwMax) {
        *(pszLine + dwLen) = c;
        if (c == '\n') {
          if (dwLen < dwMax - 1) *(pszLine + dwLen + 1) = '\0';
          break;
        }
      } else {
        break;
      }
      dwLen++;
    }
  }
  if (dwLen >= dwMax) dwErr = ERROR_MORE_DATA;
  return dwErr;

}

/*------------------------------------------------------------------------*/
DWORD
OpenFileBuffer(PFILEBUFFER pfbuff, LPCSTR filename) {
  DWORD dwErr = ERROR_SUCCESS;

  pfbuff->dwLen = 0;
  pfbuff->dwPos = 0;
  pfbuff->fEof = FALSE;
  pfbuff->hFile = CreateFile(filename, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (pfbuff->hFile == INVALID_HANDLE_VALUE) {
    dwErr = GetLastError();
  }
  return dwErr;
}

/*------------------------------------------------------------------------*/
DWORD
ReadFileChar(PFILEBUFFER pfbuff, PCHAR p) {
  DWORD dwRead;
  DWORD dwErr = ERROR_SUCCESS;

  if (pfbuff->fEof == TRUE) {
    dwErr = ERROR_HANDLE_EOF;
    goto finish_func;
  }
  if (pfbuff->dwPos < pfbuff->dwLen) {
    *p = (CHAR)*(pfbuff->buff + pfbuff->dwPos);
    pfbuff->dwPos++;
  } else {
    if (!ReadFile(pfbuff->hFile, pfbuff->buff, MAX_FILE_BUFF, &dwRead, NULL)) {
      dwErr = GetLastError();
      goto finish_func;
    } else {
      if (dwRead == 0) {
        dwErr = ERROR_HANDLE_EOF;
        pfbuff->fEof = TRUE;
        goto finish_func;
      } else {
        pfbuff->dwLen = dwRead;
        pfbuff->dwPos = 0;
        *p = (CHAR)*(pfbuff->buff + pfbuff->dwPos);
        (pfbuff->dwPos)++;
      }
    }
  }
finish_func:
  return dwErr;
}

/*------------------------------------------------------------------------*/
DWORD
ReadFileLine(PFILEBUFFER pfbuff, LPSTR pLine, DWORD dwLen) {
  LPSTR p = pLine;
  DWORD dwErr = ERROR_SUCCESS;

  while ((DWORD)(p - pLine) < dwLen - 1) {
    dwErr = ReadFileChar(pfbuff, p);
    if ((dwErr != ERROR_SUCCESS) && (dwErr != ERROR_HANDLE_EOF)) {
      goto finish_func;
    }
    if (dwErr == ERROR_HANDLE_EOF) {
      *p = '\0';
      goto finish_func;
    }
    if (*p == '\n') {
      *(p + 1) = '\0';
      goto finish_func;
    }
    p++;
  }
finish_func:
  return dwErr;
}

/*------------------------------------------------------------------------*/
DWORD
CloseFileBuffer(PFILEBUFFER pfbuff) {
  return CloseHandleClean(&(pfbuff->hFile));
}

/*------------------------------------------------------------------------*/
#define EDIT_MAX 32766

VOID
ShowCharMessage(HWND hwnd, CHAR c) {
  LPCSTR p;
  LPSTR q;
  CHAR buff[EDIT_MAX + 1];
  CHAR buff2[2] = "*";

  Edit_GetText(hwnd, buff, EDIT_MAX);

  /* GfBbgRg[̕ꍇ͐؂l */
  if (STRLEN(buff) + 4 > EDIT_MAX) {
    p = buff + (EDIT_MAX / 2);
    while(*p) {
      if (*p == '\n') break;
      p++;
    }
    if (*p) p++;
    q = buff;
    while (*p) {
      *q = *p;
      q++; p++;
    }
    *q = '\0';
  }

  buff2[0] = c;
  STRCAT(buff, buff2);
  Edit_SetText(hwnd, buff);
  SendMessage(hwnd, WM_VSCROLL, SB_BOTTOM, 0L);
  UpdateWindow(hwnd);

  return;
}

/*------------------------------------------------------------------------*/
VOID
ShowLineMessage(HWND hwnd, LPCSTR pLine) {
  LPCSTR p;
  LPSTR q, pLine2 = NULL;
  CHAR buff[EDIT_MAX + 1];
  UINT uCnt = 0;
  DWORD dwErr = ERROR_SUCCESS;

  /* pLinẻs\nꕶ\r\nɕύX */
  uCnt = 0;
  for (p = pLine; *p; p++) {
    if (*p == '\n') uCnt++;
  }
  pLine2 = (LPSTR)MALLOC(sizeof(CHAR) * (STRLEN(pLine) + uCnt + 2 + 1));
  if (!pLine2) {
    dwErr = ERROR_OUTOFMEMORY;
    goto finish_func;
  }

  q = pLine2;
  for (p = pLine; *p; p++) {
    if (*p == '\n') {
      *q = '\r';
      q++;
    }
    *q = *p;
    q++;
  }
  *q = '\0';
  if ((*pLine2) && (*(pLine2 + STRLEN(pLine2) - 1) != '\n')) {
    STRCAT(pLine2, "\r\n");
  }

  Edit_GetText(hwnd, buff, EDIT_MAX);

  /* GfBbgRg[̕ꍇ͐؂l */
  if (STRLEN(buff) + STRLEN(pLine2) + 4 > EDIT_MAX) {
    p = buff + (EDIT_MAX / 2);
    while(*p) {
      if (*p == '\n') break;
      p++;
    }
    if (*p) p++;
    q = buff;
    while (*p) {
      *q = *p;
      q++; p++;
    }
    *q = '\0';
  }

  /* Kvɉĉs} */
  if ((*buff) & (*(buff + STRLEN(buff) - 1) != '\n')) {
    STRCAT(buff, "\r\n");
  }

  STRCAT(buff, pLine2);
  Edit_SetText(hwnd, buff);
  SendMessage(hwnd, WM_VSCROLL, SB_BOTTOM, 0L);
  UpdateWindow(hwnd);

finish_func:
  if (pLine2) FREE(pLine2);

  return;
}

/*------------------------------------------------------------------------*/

