/**
 * util.h
 * Copyright (C) 2006 Kazuyoshi Kakihara
 *
 * This file is part of JanusHS.
 *
 * JanusHS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#ifndef UTIL_H_
#define UTIL_H_

#include <windows.h>
#include <Shlwapi.h>

#define CRLF "\r\n"

#define MAX_LINE 1024
#define MAX_FILE_BUFF 65536

#define MALLOC(n) HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, (n))
#define FREE(p) if ((p)) { HeapFree(GetProcessHeap(), 0, (p)); (p) = NULL; }
#define STRCAT(s1, s2) lstrcat(s1, s2)
#define STRCHR(s, i) _ustrchr((s), (i))
#define STRCPY(s1, s2) lstrcpy((s1), (s2))
#define STRDUP(s) _ustrdup((s))
#define STRLEN(s) lstrlen((s))
#define STRCMP(s1, s2) lstrcmp((s1), (s2))
#define STRNCMP(s1, s2, n) (_ustrncmp((s1), (s2), (n))) 
#define STRNCPY(s1, s2, n) (_ustrncpy((s1), (s2), (n)))
#define STRNCMPL(s1, s2) (_ustrncmp((s1), (s2), (lstrlen(s1))))

typedef struct bufflist_node {
  LPVOID buff;
  struct bufflist_node *next;
} BUFFLIST_NODE, *PBUFFLIST_NODE;

typedef struct bufflist {
  LPVOID (*clone_buff)(LPVOID src);
  VOID (*delete_buff)(LPVOID buff);
  PBUFFLIST_NODE phead;
} BUFFLIST, *PBUFFLIST;

VOID strtrim(LPSTR str);
LPSTR strgetat(LPCSTR pLine, WORD iSep, INT iCnt);
LPSTR _ustrchr(LPCSTR str, WORD w);
LPSTR _ustrdup(LPCSTR str); 
LPSTR strryen(LPCSTR str);
LPSTR _ustrncpy(LPSTR s1, LPCSTR s2, INT n);
INT _ustrncmp(LPCSTR s1, LPCSTR s2, INT n);
LPSTR _strcatmax(LPSTR s1, LPCSTR s2, INT n); 

/* Xgp֐ */
PBUFFLIST bufflist_create(LPVOID (*clone_buff)(LPVOID src), VOID (*delete_buff)(LPVOID buff));
VOID bufflist_release(PBUFFLIST plist);
DWORD bufflist_add_node_head(PBUFFLIST plist, LPVOID buff);
DWORD bufflist_add_node_tail(PBUFFLIST plist, LPVOID buff);
DWORD bufflist_add_node_sorted(PBUFFLIST plist, LPVOID buff, INT (*compare)(LPVOID pbuff1, LPVOID pbuff2));
PBUFFLIST bufflist_clone(PBUFFLIST psrc);
INT bufflist_getcount(PBUFFLIST plist);

/* ̃Xgp֐ */
#define STRLIST BUFFLIST
#define PSTRLIST PBUFFLIST
#define PSTRLIST_ITERATOR PBUFFLIST_NODE
PSTRLIST strlist_create(VOID);
VOID strlist_release(PSTRLIST plist);
LPVOID strlist_clone_buff(LPVOID src);
VOID strlist_delete_buff(LPVOID buff);
DWORD strlist_add_node(PSTRLIST plist, PCSTR str);
LPSTR strlist_get_node(PSTRLIST plist, INT i);
LPSTR strlist_concatenate(PSTRLIST plist, PCSTR pSep);
PSTRLIST_ITERATOR strlist_itr_gethead(PSTRLIST plist);
PSTRLIST_ITERATOR strlist_itr_getnext(PSTRLIST_ITERATOR pitr);
LPSTR strlist_itr_getstr(PSTRLIST_ITERATOR pitr);

/* 񑀍֐ */
DWORD strsplitgetat(LPSTR pstr, INT i, DWORD pos, LPSTR pbuff, INT max);
LPSTR escapechar(LPSTR pstr, CHAR toEscaped, CHAR escape);
LPSTR sjis2utf8(LPSTR psrc);
LPSTR utf82sjis(LPSTR psrc);
LPSTR unescapeutf82sjis(LPCSTR src);
LPSTR emailgetname(LPCSTR pemail);
LPSTR emailgetemail(LPCSTR pemail);
LPSTR emailgetcomment(LPCSTR pemail);

/* t@Cnh֐ */
typedef struct filebuffer {
  LPVOID buff[MAX_FILE_BUFF + 1];
  DWORD dwLen;
  DWORD dwPos;
  HANDLE hFile;
  BOOL fEof;
} FILEBUFFER, *PFILEBUFFER;

DWORD CreateWritePipe(PHANDLE phPWrite, PHANDLE phCRead);
DWORD CreateReadPipe(PHANDLE phPRead, PHANDLE phCWrite);
DWORD ReadPipeLine(PHANDLE phRead, PSTR pszLine, DWORD dwMax);
DWORD CloseHandleClean(PHANDLE phFile);
DWORD OpenFileBuffer(PFILEBUFFER pfbuff, PCSTR filename);
DWORD ReadFileChar(PFILEBUFFER pfbuff, PCHAR p);
DWORD ReadFileLine(PFILEBUFFER pfbuff, PSTR pLine, DWORD dwLen);
DWORD CloseFileBuffer(PFILEBUFFER pfbuff);

/* EChE֐ */
VOID ShowCharMessage(HWND hwnd, CHAR c);
VOID ShowLineMessage(HWND hwnd, PCSTR pLine);

#define DEBUGLEVEL 1
#define DEBUGERROR(c, d) \
if ((c) < DEBUGLEVEL) {\
  LPVOID lpMsgBuf;\
  CHAR szMsg[2048 + 1];\
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, \
  NULL, (d), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL); \
  wsprintf(szMsg, "%s\n%s:%d", lpMsgBuf, __FILE__, __LINE__); \
  MessageBox(NULL, szMsg, "Error", MB_OK | MB_ICONINFORMATION); \
  LocalFree(lpMsgBuf); \
}


#endif /*UTIL_H_*/
